#!/usr/bin/python
# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.

import unittest
from dougrain import curie
from dougrain import link


class CurieTest(unittest.TestCase):
    def setUp(self):
        self.curies = curie.CurieCollection()
        tlink = lambda url: link.Link(dict(href=url, templated=True),
                                      "http://localhost/api/products/1")
        self.curies['role'] = tlink("http://localhost/roles/{rel}")
        self.curies['spec'] = tlink("/specifications/{rel}")

    def testExpandsAbsoluteCurie(self):
        self.assertEquals("http://localhost/roles/host",
                          self.curies.expand("role:host"))

    def testExpandsRelativeCurie(self):
        self.assertEquals("http://localhost/specifications/color",
                          self.curies.expand("spec:color"))

    def testNullExpandsAbsoluteUrl(self):
        self.assertEquals("http://localhost/tags/foo",
                          self.curies.expand("http://localhost/tags/foo"))

    def testNullExpandsLinkWithUnknownCurieName(self):
        self.assertEquals("unknowncurie:foo",
                          self.curies.expand("unknowncurie:foo"))

    def testNullExpandsLinkWithNoCurieName(self):
        self.assertEquals("next", self.curies.expand("next"))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.

import unittest
from dougrain import link


class TestParseAbsoluteLink(unittest.TestCase):
    def setUp(self):
        self.link = link.Link(
            {'href': "http://localhost/"},
            None)

    def testLoadsAbsoluteHref(self):
        self.assertEquals("http://localhost/", self.link.href)

    def testCalculatesUrl(self):
        self.assertEquals("http://localhost/", self.link.url())

    def testNameIsMissing(self):
        self.assertFalse(hasattr(self.link, 'name'))

    def testTitleIsMissing(self):
        self.assertFalse(hasattr(self.link, 'title'))


class TestParseRelativeLink(unittest.TestCase):
    def setUp(self):
        self.link = link.Link(
            {'href': "/foo"},
            "http://localhost/")

    def testLoadsAbsoluteHref(self):
        self.assertEquals("/foo", self.link.href)

    def testCalculatesUrl(self):
        self.assertEquals("http://localhost/foo", self.link.url())

    def testNameIsMissing(self):
        self.assertFalse(hasattr(self.link, 'name'))

    def testTitleIsMissing(self):
        self.assertFalse(hasattr(self.link, 'title'))


class TestParseAdditionalAttributes(unittest.TestCase):
    def setUp(self):
        self.link = link.Link(
            {
                'href': "/foo",
                'name': "bar",
                'title': "Bar",
                'type': "application/hal+json",
                'profile': "/profiles/foo",
                'hreflang': "en",
                'deprecation': "/deprecation/foo"
            },
            "http://localhost/"
        )

    def testLoadsName(self):
        self.assertEquals("bar", self.link.name)

    def testLoadsTitle(self):
        self.assertEquals("Bar", self.link.title)

    def testLoadsType(self):
        self.assertEquals("application/hal+json", self.link.type)

    def testLoadsProfile(self):
        self.assertEquals("/profiles/foo", self.link.profile)

    def testLoadsHreflang(self):
        self.assertEquals("en", self.link.hreflang)

    def testLoadsDeprecation(self):
        self.assertEquals("/deprecation/foo", self.link.deprecation)


class TestExpandTemplatedLink(unittest.TestCase):
    def setUp(self):
        self.link = link.Link(
            {
                'href': "/foo/{arg1}",
                'templated': True
            },
            "http://localhost/"
        )

    def testHasArgs(self):
        self.assertEquals(["arg1"], self.link.variables)

    def testSubstituteNoArgs(self):
        self.assertEquals("http://localhost/foo/", self.link.url())

    def testSubstituteArg(self):
        self.assertEquals("http://localhost/foo/1-bar",
                          self.link.url(arg1="1-bar"))

    def testPreservesTemplate(self):
        self.assertEquals("http://localhost/foo/{arg1}", self.link.template)


class TestDoNotExpandNonTemplateLinks(unittest.TestCase):
    def setUp(self):
        self.link = link.Link(
            {
                'href': "/foo/{arg1}"
            },
            "http://localhost/"
        )

    def testHasNoArgs(self):
        self.assertEquals([], self.link.variables)

    def testSubstituteNoArgs(self):
        self.assertEquals("http://localhost/foo/{arg1}", self.link.url())

    def testSubstituteArg(self):
        self.assertEquals("http://localhost/foo/{arg1}",
                          self.link.url(arg1="1-bar"))

    def testPreservesTemplate(self):
        self.assertEquals("http://localhost/foo/{arg1}", self.link.template)


class TestExtractVariablesFromLink(unittest.TestCase):
    def assertVariables(self, variables, href):
        if isinstance(variables, str):
            variables = variables.split()

        lnk = link.Link(
            {
                'href': href,
                'templated': True
            },
            "http://localhost/"
        )
        self.assertEquals(variables, lnk.variables)

    def testNoVariables(self):
        self.assertVariables([], "/index.html")

    def testSingleVariableLevel1(self):
        self.assertVariables('var', "{var}")
        self.assertVariables('hello', "{hello}")

    def testUrlEncodeVariableLevel2(self):
        self.assertVariables("var", "{+var}")
        self.assertVariables("hello", "{+hello}")
        self.assertVariables("path", "{+path}/here")
        self.assertVariables("path", "here?ref={+path}")

    def testMultipleVariablesLevel3(self):
        self.assertVariables("x y", "map?{x,y}")
        self.assertVariables("x hello y", "{x,hello,y}")
        self.assertVariables("x hello y", "{+x,hello,y}")
        self.assertVariables("path x", "{+path,x}/here")
        self.assertVariables("x hello y", "{#x,hello,y}")
        self.assertVariables("path x", "{#path,x}/here")
        self.assertVariables("var", "X{.var}")
        self.assertVariables("x y", "X{.x,y}")
        self.assertVariables("var", "{/var}")
        self.assertVariables("var x", "{/var,x}/here")
        self.assertVariables("x y", "{;x,y}")
        self.assertVariables("x y empty", "{;x,y,empty}")
        self.assertVariables("x y", "{?x,y}")
        self.assertVariables("x y empty", "{?x,y,empty}")
        self.assertVariables("x", "?fixed=yes{&x}")
        self.assertVariables("x y empty", "{&x,y,empty}")

    def testLevel4(self):
        self.assertVariables("var", "{var:3}")
        self.assertVariables("var", "{var:30}")
        self.assertVariables("list", "{list}")
        self.assertVariables("list", "{list*}")
        self.assertVariables("keys", "{keys}")
        self.assertVariables("keys", "{keys*}")
        self.assertVariables("path", "{+path:6}/here")
        self.assertVariables("list", "{+list}")
        self.assertVariables("list", "{+list*}")
        self.assertVariables("keys", "{+keys}")
        self.assertVariables("keys", "{+keys*}")
        self.assertVariables("path", "{#path:6}/here")
        self.assertVariables("list", "{#list}")
        self.assertVariables("list", "{#list*}")
        self.assertVariables("keys", "{#keys}")
        self.assertVariables("keys", "{#keys*}")
        self.assertVariables("var", "X{.var:3}")
        self.assertVariables("list", "X{.list}")
        self.assertVariables("list", "X{.list*}")
        self.assertVariables("keys", "X{.keys}")
        self.assertVariables("var", "{/var:1,var}")
        self.assertVariables("list", "{/list}")
        self.assertVariables("list", "{/list*}")
        self.assertVariables("list path", "{/list*,path:4}")
        self.assertVariables("keys", "{/keys}")
        self.assertVariables("keys", "{/keys*}")
        self.assertVariables("hello", "{;hello:5}")
        self.assertVariables("list", "{;list}")
        self.assertVariables("list", "{;list*}")
        self.assertVariables("keys", "{;keys}")
        self.assertVariables("keys", "{;keys*}")
        self.assertVariables("var", "{?var:3}")
        self.assertVariables("list", "{?list}")
        self.assertVariables("list", "{?list*}")
        self.assertVariables("keys", "{?keys}")
        self.assertVariables("keys", "{?keys*}")
        self.assertVariables("var", "{&var:3}")
        self.assertVariables("list", "{&list}")
        self.assertVariables("list", "{&list*}")
        self.assertVariables("keys", "{&keys}")
        self.assertVariables("keys", "{&keys*}")


class TestIteration(unittest.TestCase):
    def testASingleLinkCanBeIterated(self):
        the_link = link.Link({"href": "/"}, "http://localhost/")
        count = 0
        for a_link in the_link:
            count += 1
            self.assertEquals(the_link, a_link)

        self.assertEquals(1, count)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.

import unittest
import dougrain


class ParseSimpleTestMixin(object):
    def setUp(self):
        self.doc = dougrain.Document.from_object({"name": "David Bowman"},
                                                 draft=self.DRAFT)

    def testParseSimple(self):
        self.assertEquals(self.doc.properties['name'], "David Bowman")

    def testHasEmptyLinks(self):
        self.assertEquals(self.doc.links, {})


class ParseSimpleTestDraft5(ParseSimpleTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5


class ParseSimpleTestDraft4(ParseSimpleTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4


class ParseSimpleTestDraft3(ParseSimpleTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3


#


class ParseLinksTestMixin(object):
    OBJECT = {
        "_links": {
            "self": {"href": "dougrain"},
            "next": {
                "href": "http://localhost/wharris/esmre",
                "title": "Next"
            },
            "parent": {"href": "/wharris/"},
            "images": [
                {"href": "/foo"},
                {"href": "/bar"}
            ],
        }
    }

    def setUp(self):
        self.doc = dougrain.Document.from_object(
            self.OBJECT,
            base_uri="http://localhost/wharris/dougrain",
            draft=self.DRAFT)

    def testLoadsSingleLinkHref(self):
        self.assertEquals("http://localhost/wharris/esmre",
                          self.doc.links["next"].href)

    def testLoadsSingleLinkURL(self):
        self.assertEquals("http://localhost/wharris/esmre",
                          self.doc.links["next"].url())

    def testLoadsRelativeURLHref(self):
        self.assertEquals("/wharris/",
                          self.doc.links["parent"].href)

    def testAbsolutesRelativeURL(self):
        self.assertEquals("http://localhost/wharris/",
                          self.doc.links["parent"].url())

    def testLoadsSelfAsLinkAndAttribute(self):
        self.assertEquals("dougrain", self.doc.links["self"].href)
        self.assertEquals("http://localhost/wharris/dougrain",
                          self.doc.links["self"].url())

        self.assertEquals(self.doc.links["self"].url(), self.doc.url())

    def testLoadsTitle(self):
        self.assertEquals("Next", self.doc.links["next"].title)
        self.assertFalse(hasattr(self.doc.links["parent"], "title"))
        self.assertFalse(hasattr(self.doc.links["self"], "title"))

    def testLoadsArrayOfLinks(self):
        self.assertEquals(["/foo", "/bar"],
                          [link.href for link in self.doc.links['images']])

    def testLinksIsNotAnAttribute(self):
        self.assertFalse('_links' in self.doc.properties)
        self.assertFalse(hasattr(self.doc, '_links'))


class ParseLinksTestDraft5(ParseLinksTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5


class ParseLinksTestDraft4(ParseLinksTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4


class ParseLinksTestDraft3(ParseLinksTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3


#


class ParseEmbeddedObjectsTestMixin(object):
    def setUp(self):
        self.doc = dougrain.Document.from_object(
            self.OBJECT,
            base_uri="http://localhost/people/",
            draft=self.DRAFT)

    def testLoadsSingleEmbeddedObject(self):
        foo = self.doc.embedded["foo"]
        self.assertEquals("Foo", foo.properties['name'])
        self.assertEquals(88888888, foo.properties['size'])

    def testLoadsArrayOfEmbeddedObjects(self):
        self.assertEquals(["Bar 1", "Bar 2"],
                          [bar.properties['title']
                           for bar in self.doc.embedded['bar']])

    def testLoadsLinksInEmbeddedObject(self):
        link = self.doc.embedded["bundy"].links["next"]
        self.assertEquals("/people/2", link.href)
        self.assertEquals("http://localhost/people/2", link.url())

    def testEmbeddedIsNotAnAttribute(self):
        self.assertFalse('_embedded' in self.doc.properties)
        self.assertFalse(hasattr(self.doc, '_embedded'))


class ParseEmbeddedObjectsTestDraft5(ParseEmbeddedObjectsTestMixin,
                                     unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5
    OBJECT = {
        "_links": {
            "foo": {'href': "/foo"},
            "bar": [{'href': "/bar/1"}, {'href': "/bar/2"}],
            "bundy": {'href': "/bundy"}
        },
        "_embedded": {
            "foo": {
                "_links": {"self": {'href': "/foo"}},
                "name": "Foo",
                "size": 88888888
            },
            "bar": [
                {
                    "_links": {"self": {'href': "/bar/1"}},
                    "title": "Bar 1"
                },
                {
                    "_links": {"self": {'href': "/bar/2"}},
                    "title": "Bar 2"
                }
            ],
            "bundy": {
                "_links": {
                    "self": {'href': "/bundy"},
                    "next": {"href": "/people/2"}
                }
            }
        }
    }


class ParseEmbeddedObjectsTestDraft4(ParseEmbeddedObjectsTestMixin,
                                     unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4
    OBJECT = {
        "_embedded": {
            "foo": {
                "_links": {"self": {'href': "/foo"}},
                "name": "Foo",
                "size": 88888888
            },
            "bar": [
                {
                    "_links": {"self": {'href': "/bar/1"}},
                    "title": "Bar 1"
                },
                {
                    "_links": {"self": {'href': "/bar/2"}},
                    "title": "Bar 2"
                }
            ],
            "bundy": {
                "_links": {
                    "self": {'href': "/bundy"},
                    "next": {"href": "/people/2"}
                }
            }
        }
    }


class ParseEmbeddedObjectsTestDraft3(ParseEmbeddedObjectsTestMixin,
                                     unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5
    OBJECT = ParseEmbeddedObjectsTestDraft4.OBJECT


#


class CurieExpansionTestMixin(object):

    def setUp(self):
        self.doc = dougrain.Document.from_object(self.OBJECT,
                                                 draft=self.DRAFT)

    def testExposesCurieCollection(self):
        expanded = self.doc.expand_curie('role:category')
        self.assertEquals("http://localhost/roles/category",
                          expanded)

    def testEmbeddedObjectHasParentCuries(self):
        sizing_doc = self.doc.embedded['role:sizing']
        self.assertEquals("http://localhost/roles/host",
                          sizing_doc.expand_curie('role:host'))

    def testParentObjectDoesNotHaveEmbeddedCuried(self):
        self.assertEquals('dim:weight', self.doc.expand_curie('dim:weight'))

    def testEmbeddedObjectExtendsParentCuries(self):
        sizing_doc = self.doc.embedded['role:sizing']
        self.assertEquals("http://localhost/dimension/weight",
                          sizing_doc.expand_curie('dim:weight'))

    def testEmbeddedObjectCurieOverridesParentCurie(self):
        coloring_doc = self.doc.embedded['role:coloring']
        self.assertEquals("http://localhost/imagefiles/photo",
                          coloring_doc.expand_curie('image:photo'))


class CurieExpansionTestDraft5(CurieExpansionTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5
    OBJECT = {
        '_links': {
            'curies': [
                {
                    'href': "http://localhost/roles/{rel}",
                    'name': 'role',
                    'templated': True
                },
                {
                    'href': "http://localhost/images/{rel}",
                    'name': 'image',
                    'templated': True
                }
            ],
            'role:host': {'href': "/hosts/1"},
            'role:sizing': {'href': "http://localhost/sizing/1"},
            'role:coloring': {'href': "http://localhost/coloring/1"}
        },
        '_embedded': {
            'role:sizing': {
                '_links': {
                    'self': {'href': "http://localhost/sizing/1"},
                    'curies': [
                        {
                            'href': "http://localhost/dimension/{rel}",
                            'name': 'dim',
                            'templated': True
                        }
                    ]
                }
            },
            'role:coloring': {
                '_links': {
                    'self': {'href': "http://localhost/coloring/1"},
                    'curies': [
                        {
                            'href':
                            "http://localhost/imagefiles/{rel}",
                            'name': 'image',
                            'templated': True
                        }
                    ]
                }
            },
        }
    }


class CurieExpansionTestDraft4(CurieExpansionTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4
    OBJECT = {
        '_links': {
            'curies': [
                {
                    'href': "http://localhost/roles/{rel}",
                    'name': 'role',
                    'templated': True
                },
                {
                    'href': "http://localhost/images/{rel}",
                    'name': 'image',
                    'templated': True
                }
            ],
            'role:host': {'href': "/hosts/1"}
        },
        '_embedded': {
            'role:sizing': {
                '_links': {
                    'self': {'href': "http://localhost/sizing/1"},
                    'curies': [
                        {
                            'href': "http://localhost/dimension/{rel}",
                            'name': 'dim',
                            'templated': True
                        }
                    ]
                }
            },
            'role:coloring': {
                '_links': {
                    'self': {'href': "http://localhost/coloring/1"},
                    'curies': [
                        {
                            'href':
                            "http://localhost/imagefiles/{rel}",
                            'name': 'image',
                            'templated': True
                        }
                    ]
                }
            },
        }
    }


class CurieExpansionTestDraft3(CurieExpansionTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3

    OBJECT = {
        '_links': {
            'curie': [
                {
                    'href': "http://localhost/roles/{rel}",
                    'name': 'role',
                    'templated': True
                },
                {
                    'href': "http://localhost/images/{rel}",
                    'name': 'image',
                    'templated': True
                }
            ],
            'role:host': {'href': "/hosts/1"}
        },
        '_embedded': {
            'role:sizing': {
                '_links': {
                    'curie': {
                        'href': "http://localhost/dimension/{rel}",
                        'name': 'dim',
                        'templated': True
                    }
                }
            },
            'role:coloring': {
                '_links': {
                    'curie': {
                        'href': "http://localhost/imagefiles/{rel}",
                        'name': 'image',
                        'templated': True
                    }
                }
            },
        }
    }


#


class RelsTestMixin(object):
    def setUp(self):
        self.doc = dougrain.Document(self.OBJECT, "http://localhost",
                                     draft=self.DRAFT)

    def testHasHostLinkRel(self):
        host_role = self.doc.expand_curie('role:host')
        self.assertTrue(host_role in self.doc.rels)
        self.assertEquals(["http://localhost/hosts/1"],
                          [x.url() for x in self.doc.rels[host_role]])

    def testHasConsumerEmbeddedRel(self):
        consumer_role = self.doc.expand_curie('role:consumer')
        self.assertTrue(consumer_role in self.doc.rels)
        consumer = self.doc.rels[consumer_role][0]
        self.assertEquals("http://localhost/clients/1", consumer.url())
        self.assertEquals("Client 1", consumer.properties['name'])

    def testHasApplicationLinkAndEmbeddedRels(self):
        application_role = self.doc.expand_curie('role:application')
        applications = self.doc.rels[application_role]
        self.assertTrue(self.doc.embedded['role:application'] in applications)
        self.assertTrue(self.doc.links['role:application'] in applications)

    def testEmbeddedRelOverridesLinkRelWithSameHref(self):
        dept_role = self.doc.expand_curie('role:dept')

        filter_url = "http://localhost/departments/2"
        dept_2_link = [link for link in self.doc.links['role:dept']
                       if link.url() == filter_url][0]
        dept_2_embedded = [link for link in self.doc.embedded['role:dept']
                           if link.url() == filter_url][0]

        departments = self.doc.rels[dept_role]
        self.assertTrue(dept_2_embedded in departments)
        self.assertFalse(dept_2_link in departments)

        urls = [x.url() for x in departments]
        urls.sort()
        self.assertEquals(
            ["http://localhost/departments/%d" % x for x in [1, 2, 3]],
            urls)


class RelsTestDraft5(RelsTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5
    OBJECT = {
        '_links': {
            'curies': [
                {
                    'href': "/roles/{rel}",
                    'name': 'role',
                    'templated': True
                },
                {
                    'href': "http://localhost/images/{rel}",
                    'name': 'image',
                    'templated': True
                }
            ],
            'role:host': {'href': "/hosts/1"},
            'role:application': {'href': "/apps/1"},
            'role:dept': [
                {'href': "/departments/1"},
                {'href': "/departments/2"}
            ]
        },
        '_embedded': {
            'role:consumer': {
                '_links': {
                    'self': {
                        'href': '/clients/1'
                    }
                },
                'name': "Client 1"
            },
            'role:application': {
                '_links': {
                    'self': {'href': "/apps/2"}
                }
            },
            'role:dept': [
                {
                    '_links': {
                        'self': {
                            'href': "http://localhost/departments/2"
                        }
                    },
                },
                {
                    '_links': {
                        'self': {'href': "/departments/3"}
                    }
                }
            ]
        }
    }


class RelsTestDraft4(RelsTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4
    OBJECT = RelsTestDraft5.OBJECT


class RelsTestDraft3(RelsTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3
    OBJECT = {
        '_links': {
            'curie': [
                {
                    'href': "/roles/{rel}",
                    'name': 'role',
                    'templated': True
                },
                {
                    'href': "http://localhost/images/{rel}",
                    'name': 'image',
                    'templated': True
                }
            ],
            'role:host': {'href': "/hosts/1"},
            'role:application': {'href': "/apps/1"},
            'role:dept': [
                {'href': "/departments/1"},
                {'href': "/departments/2"}
            ]
        },
        '_embedded': {
            'role:consumer': {
                '_links': {
                    'self': {
                        'href': '/clients/1'
                    }
                },
                'name': "Client 1"
            },
            'role:application': {
                '_links': {
                    'self': {'href': "/apps/2"}
                }
            },
            'role:dept': [
                {
                    '_links': {
                        'self': {
                            'href': "http://localhost/departments/2"
                        }
                    },
                },
                {
                    '_links': {
                        'self': {'href': "/departments/3"}
                    }
                }
            ]
        }
    }


#


class SerializeTestsMixin(object):
    def checkEqualObjects(self, obj):
        doc = dougrain.Document.from_object(obj, "http://localhost",
                                            draft=self.DRAFT)
        self.assertEquals(obj, doc.as_object())

    def testSimple(self):
        self.checkEqualObjects({})

    def testAttributes(self):
        self.checkEqualObjects({"latlng": [53.0, -0.001],
                                "altitude": 10.0,
                                "haccuracy": 5.0,
                                "vacuracy": 10.0})

    def testLinks(self):
        self.checkEqualObjects(ParseLinksTestMixin.OBJECT)

    def testEmbeddedObjects(self):
        self.checkEqualObjects(self.EMBEDDED_CASE)

    def testRels(self):
        self.checkEqualObjects(self.RELS_CASE)

    def testCuries(self):
        self.checkEqualObjects(self.CURIES_CASE)


class SerializeTestsDraft5(SerializeTestsMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5
    EMBEDDED_CASE = ParseEmbeddedObjectsTestDraft5.OBJECT
    RELS_CASE = RelsTestDraft5.OBJECT
    CURIES_CASE = CurieExpansionTestDraft5.OBJECT


class SerializeTestsDraft4(SerializeTestsMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4
    EMBEDDED_CASE = ParseEmbeddedObjectsTestDraft4.OBJECT
    RELS_CASE = RelsTestDraft4.OBJECT
    CURIES_CASE = CurieExpansionTestDraft4.OBJECT


class SerializeTestsDraft3(SerializeTestsMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3
    EMBEDDED_CASE = ParseEmbeddedObjectsTestDraft3.OBJECT
    RELS_CASE = RelsTestDraft3.OBJECT
    CURIES_CASE = CurieExpansionTestDraft3.OBJECT


#


class AttributeMutationTests(unittest.TestCase):
    def testSetAttributeAddsAttribute(self):
        doc = dougrain.Document.empty()
        doc.set_property('foo', "bar")

        self.assertEquals("bar", doc.properties['foo'])

    def testSetAttributeUpdatesAttribute(self):
        doc = dougrain.Document.empty()
        doc.set_property('foo', "bar")

        doc.set_property('foo', "bundy")

        self.assertEquals("bundy", doc.properties['foo'])

    def testRemoveAttributeRemovesAttribute(self):
        doc = dougrain.Document.empty()
        doc.set_property('foo', "bar")

        doc.delete_property('foo')

        self.assertFalse(hasattr(doc, 'foo'))
        self.assertFalse('foo' in doc.properties)

    def testBuildObjectFromEmpty(self):
        target = {"latlng": [53.0, -0.001],
                  "altitude": 10.0,
                  "haccuracy": 5.0,
                  "vacuracy": 10.0}
        target_doc = dougrain.Document.from_object(target)

        doc = dougrain.Document.empty()

        for key, value in target_doc.properties.iteritems():
            doc.set_property(key, value)

        self.assertEquals(target_doc.as_object(), doc.as_object())


#


class AddLinkStringTests(unittest.TestCase):
    def add_link(self, doc, rel, href, **kwargs):
        doc.add_link(rel, href, **kwargs)

    def testAddSimpleLink(self):
        target = {
            '_links': {
                'self': {'href': "http://localhost/1"}
            }
        }

        target_doc = dougrain.Document.from_object(target, "http://localhost/")

        doc = dougrain.Document.empty()
        self.add_link(doc, 'self', "http://localhost/1")

        self.assertEquals(target, doc.as_object())
        self.assertEquals(target_doc.links, doc.links)

    def testAddWrappedLink(self):
        target = {
            '_links': {
                'self': {'href': "http://localhost/1"},
                'item': [{'href': "http://localhost/1/1"}]
            }
        }

        target_doc = dougrain.Document.from_object(target, "http://localhost/")

        doc = dougrain.Document.empty()
        self.add_link(doc, 'self', "http://localhost/1")
        self.add_link(doc, 'item', "http://localhost/1/1", wrap=list)

        self.assertEquals(target, doc.as_object())

    def testAddLinkForSecondRelKeepsFirstRel(self):
        target = {
            '_links': {
                'self': {'href': "http://localhost/1"},
                'child': {'href': "http://localhost/1/1"}
            }
        }

        target_doc = dougrain.Document.from_object(target, "http://localhost/")

        doc = dougrain.Document.empty()
        self.add_link(doc, 'self', "http://localhost/1")
        self.add_link(doc, 'child', "http://localhost/1/1")

        self.assertEquals(target, doc.as_object())
        self.assertEquals(target_doc.links, doc.links)

    def testAddSecondLinkForSameRel(self):
        target = {
            '_links': {
                'self': {'href': "http://localhost/2"},
                'child': [{'href': "http://localhost/2/1"},
                          {'href': "http://localhost/2/2"}]
            }
        }

        target_doc = dougrain.Document.from_object(target, "http://localhost/")

        doc = dougrain.Document.empty()
        self.add_link(doc, 'self', "http://localhost/2")
        self.add_link(doc, 'child', "http://localhost/2/1")
        self.add_link(doc, 'child', "http://localhost/2/2")

        self.assertEquals(target, doc.as_object())
        self.assertEquals(target_doc.links, doc.links)

    def testAddThirdLinkForSameRel(self):
        target = {
            '_links': {
                'self': {'href': "http://localhost/2"},
                'child': [{'href': "http://localhost/2/1"},
                          {'href': "http://localhost/2/2"},
                          {'href': "http://localhost/2/3"}]
            }
        }

        target_doc = dougrain.Document.from_object(target, "http://localhost/")

        doc = dougrain.Document.empty()
        self.add_link(doc, 'self', "http://localhost/2")
        self.add_link(doc, 'child', "http://localhost/2/1")
        self.add_link(doc, 'child', "http://localhost/2/2")
        self.add_link(doc, 'child', "http://localhost/2/3")

        self.assertEquals(target, doc.as_object())

        self.assertEquals(target, doc.as_object())
        self.assertEquals(target_doc.links, doc.links)

    def testAddLinkWithExtraAttributes(self):
        target = {
            '_links': {
                'self': {'href': "http://localhost/2"},
                'child': [{'href': "http://localhost/2/1",
                           'title': "First Child"},
                          {'href': "http://localhost/2/2",
                           'title': "Second Child"}]
            }
        }
        doc = dougrain.Document.empty()
        self.add_link(doc, 'self', "http://localhost/2")
        self.add_link(doc, 'child', "http://localhost/2/1",
                      title="First Child")
        self.add_link(doc, 'child', "http://localhost/2/2",
                      title="Second Child")

        self.assertEquals(target, doc.as_object())


#


class AddObjectLinkTests(AddLinkStringTests):
    def add_link(self, doc, rel, href, wrap=False, **kwargs):
        link = doc.link(href, **kwargs)
        doc.add_link(rel, link, wrap=wrap)


#


class AddDocumentLinkTests(AddLinkStringTests):
    def add_link(self, doc, rel, href, wrap=False, **kwargs):
        target = dougrain.Document.empty(href)
        self_link = target.link(href, **kwargs)
        target.add_link('self', self_link)
        doc.add_link(rel, target, wrap=wrap)

#


class AddBuilderLinkTests(AddLinkStringTests):
    def add_link(self, doc, rel, href, wrap=False, **kwargs):
        target = dougrain.Builder(href, **kwargs)
        doc.add_link(rel, target, wrap=wrap)


#


class DeleteLinkTests(unittest.TestCase):
    def testDeleteOnlyLinkForRel(self):
        initial = {
            '_links': {
                'self': {'href': "http://localhost/2"},
                'child': {'href': "http://localhost/2/1"}
            }
        }

        target = {
            '_links': {
                'self': {'href': "http://localhost/2"},
            }
        }

        doc = dougrain.Document.from_object(initial, "http://localhost/")
        target_doc = dougrain.Document.from_object(target, "http://localhost/")

        doc.delete_link("child")

        self.assertEquals(target_doc, doc)
        self.assertFalse('child' in doc.links)

    def testDeleteEveryLinkForRel(self):
        initial = {
            '_links': {
                'self': {'href': "http://localhost/2"},
                'child': [
                    {'href': "http://localhost/2/1"},
                    {'href': "http://localhost/2/2"},
                    {'href': "http://localhost/2/3"}
                ]
            }
        }

        target = {
            '_links': {
                'self': {'href': "http://localhost/2"}
            }
        }

        doc = dougrain.Document.from_object(initial, "http://localhost/")
        target_doc = dougrain.Document.from_object(target, "http://localhost/")

        doc.delete_link("child")

        self.assertEquals(target_doc.as_object(), doc.as_object())
        self.assertFalse('child' in doc.links)

    def testDeleteLastLink(self):
        initial = {
            '_links': {
                'self': {'href': "http://localhost/2"},
                'child': {'href': "http://localhost/2/1"}
            }
        }

        target = {}

        doc = dougrain.Document.from_object(initial, "http://localhost/")
        target_doc = dougrain.Document.from_object(target, "http://localhost/")

        doc.delete_link("child")
        doc.delete_link("self")

        self.assertEquals(target_doc.as_object(), doc.as_object())

    def testDeleteIndividualLinks(self):
        initial = {
            '_links': {
                'self': {'href': "http://localhost/2"},
                'child': [
                    {'href': "http://localhost/2/1"},
                    {'href': "http://localhost/2/2"},
                    {'href': "http://localhost/2/3"}
                ]
            }
        }

        doc = dougrain.Document.from_object(initial, "http://localhost/")

        doc.delete_link("child", "http://localhost/2/1")
        self.assertEquals([{'href': "http://localhost/2/2"},
                           {'href': "http://localhost/2/3"}],
                          doc.as_object()['_links']['child'])

        doc.delete_link("child", "http://localhost/2/3")
        self.assertEquals({'href': "http://localhost/2/2"},
                          doc.as_object()['_links']['child'])

        doc.delete_link("child", "http://localhost/2/2")
        self.assertFalse("child" in doc.as_object()['_links'])

        doc.delete_link("self", "http://localhost/2")
        self.assertFalse('_links' in doc.as_object())

    def testDeleteLinksWithoutRel(self):
        initial = {
            '_links': {
                'self': {'href': "http://localhost/3"},
                'child': [{'href': "http://localhost/3/1"},
                          {'href': "http://localhost/3/2"}],
                'favorite': {'href': "http://localhost/3/1"}
            }
        }

        target = {
            '_links': {
                'self': {'href': "http://localhost/3"},
                'child': {'href': "http://localhost/3/2"}
            }
        }

        doc = dougrain.Document.from_object(initial, "http://localhost/")

        doc.delete_link(href="http://localhost/3/1")

        self.assertEquals(target, doc.as_object())

    def testDeleteAllLinks(self):
        initial = {
            '_links': {
                'self': {'href': "http://localhost/3"},
                'child': [{'href': "http://localhost/3/1"},
                          {'href': "http://localhost/3/2"}],
                'favorite': {'href': "http://localhost/3/1"}
            }
        }

        target = {}

        doc = dougrain.Document.from_object(initial, "http://localhost/")

        doc.delete_link()

        self.assertEquals(target, doc.as_object())

    def testDeleteLinkWithNoMatchingRel(self):
        initial = {
            '_links': {
                'self': {'href': "http://localhost/3"},
                'child': [{'href': "http://localhost/3/1"},
                          {'href': "http://localhost/3/2"}],
                'favorite': {'href': "http://localhost/3/1"}
            }
        }

        target = {
            '_links': {
                'self': {'href': "http://localhost/3"},
                'child': [{'href': "http://localhost/3/1"},
                          {'href': "http://localhost/3/2"}],
                'favorite': {'href': "http://localhost/3/1"}
            }
        }

        doc = dougrain.Document.from_object(initial, "http://localhost/")

        doc.delete_link(rel="".join(doc.links.keys()))

        self.assertEquals(target, doc.as_object())

    def testDeleteLinkWithNoLinks(self):
        doc = dougrain.Document.empty("http://localhost/3")
        target = dougrain.Document.empty("http://localhost/3")

        doc.delete_link()

        self.assertEquals(target, doc)


#


class EmbedTestMixin(object):
    OBJECT_WITH_SELF = {
        "ham": "beans",
        "_links": {"self": {"href": "/test"}}
    }

    def setUp(self):
        self.doc = dougrain.Document.empty("http://localhost/",
                                           draft=self.DRAFT)
        self.embedded1 = dougrain.Document.from_object({"foo": "bar"},
                                                       draft=self.DRAFT)
        self.embedded2 = dougrain.Document.from_object({"spam": "eggs"},
                                                       draft=self.DRAFT)
        self.embedded3 = dougrain.Document.from_object({"ham": "beans"},
                                                       draft=self.DRAFT)
        self.embedded_with_self = dougrain.Document.from_object(
            self.OBJECT_WITH_SELF,
            draft=self.DRAFT)

    def testEmbed(self):
        expected = {
            '_embedded': {
                'child': self.embedded1.as_object()
            }
        }
        self.doc.embed('child', self.embedded1)
        self.assertEquals(expected, self.doc.as_object())
        self.assertEquals(self.embedded1, self.doc.embedded['child'])

    def testEmbedWrapped(self):
        expected = {
            '_embedded': {
                'child': [self.embedded1.as_object()]
            }
        }
        self.doc.embed('child', self.embedded1, wrap=True)
        self.assertEquals(expected, self.doc.as_object())
        self.assertEquals([self.embedded1], self.doc.embedded['child'])

    def testEmbedAnotherRel(self):
        expected = {
            '_embedded': {
                'prev': self.embedded2.as_object(),
                'next': self.embedded3.as_object(),
            }
        }
        self.doc.embed('prev', self.embedded2)
        self.doc.embed('next', self.embedded3)

        self.assertEquals(expected, self.doc.as_object())
        self.assertEquals(self.embedded2, self.doc.embedded['prev'])
        self.assertEquals(self.embedded3, self.doc.embedded['next'])

    def testEmbedSameRel(self):
        expected = {
            '_embedded': {
                'item': [
                    self.embedded1.as_object(),
                    self.embedded2.as_object(),
                    self.embedded3.as_object()
                ]
            }
        }
        self.doc.embed('item', self.embedded1)
        self.doc.embed('item', self.embedded2)
        self.doc.embed('item', self.embedded3)

        self.assertEquals(expected, self.doc.as_object())
        self.assertEquals([self.embedded1, self.embedded2, self.embedded3],
                          self.doc.embedded['item'])

    def testEmbedDocumentWithSelfLink(self):
        self.doc.embed('item', self.embedded_with_self)
        self.assertEquals(self.EXPECTED_WITH_SELF_LINK,
                          self.doc.as_object())


class EmbedTestDraft5(EmbedTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5
    EXPECTED_WITH_SELF_LINK = {
        '_links': {
            'item': {'href': "/test"}
        },
        '_embedded': {
            'item': EmbedTestMixin.OBJECT_WITH_SELF
        }
    }

    EXPECTED_WITH_SELF_LINK_WRAPPED = {
        '_links': {
            'item': [{'href': "/test"}]
        },
        '_embedded': {
            'item': [EmbedTestMixin.OBJECT_WITH_SELF]
        }
    }

    def testEmbedWrappedWithSelfLink(self):
        self.doc.embed('item', self.embedded_with_self, wrap=True)
        self.assertEquals(self.EXPECTED_WITH_SELF_LINK_WRAPPED,
                          self.doc.as_object())


class EmbedTestDraft4(EmbedTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4
    EXPECTED_WITH_SELF_LINK = {
        '_embedded': {
            'item': EmbedTestMixin.OBJECT_WITH_SELF
        }
    }


class EmbedTestDraft3(EmbedTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3
    EXPECTED_WITH_SELF_LINK = {
        '_embedded': {
            'item': EmbedTestMixin.OBJECT_WITH_SELF
        }
    }


#


class TestIteration(unittest.TestCase):
    def testASingleDocumentCanBeIterated(self):
        for draft in (dougrain.drafts.DRAFT_3,
                      dougrain.drafts.DRAFT_4,
                      dougrain.drafts.DRAFT_5):
            the_doc = dougrain.Document.empty("http://localhost/1",
                                              draft=draft)
            the_doc.add_link('self', "http://localhost/1")

            count = 0
            for a_doc in the_doc:
                count += 1
                self.assertEquals(the_doc, a_doc)

            self.assertEquals(1, count)


#


class DeleteEmbeddedTestsMixin(object):
    EMBEDDED_KEY = dougrain.document.EMBEDDED_KEY

    def make_doc(self, href, *args, **kwargs):
        result = dougrain.Document.empty("http://localhost", *args,
                                         draft=self.DRAFT, **kwargs)
        result.add_link('self', href)
        return result

    def testDeleteOnlyEmbedForRel(self):
        doc = self.make_doc("http://localhost/2")
        doc.embed('child', self.make_doc("http://localhost/2/1"))
        doc.embed('root', self.make_doc("http://localhost/"))

        target_doc = self.make_doc("http://localhost/2")
        target_doc.embed('root', self.make_doc("http://localhost/"))

        doc.delete_embedded("child")

        self.assertEquals(target_doc.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))
        self.assertFalse('child' in doc.embedded)
        self.assertTrue('root' in doc.embedded)

    def testDeleteEveryEmbedForRel(self):
        doc = self.make_doc("http://localhost/2")
        doc.embed('root', self.make_doc("http://localhost/"))
        doc.embed('child', self.make_doc("http://localhost/2/1"))
        doc.embed('child', self.make_doc("http://localhost/2/1"))
        doc.embed('child', self.make_doc("http://localhost/2/1"))

        target_doc = self.make_doc("http://localhost/2")
        target_doc.embed('root', self.make_doc("http://localhost/"))

        doc.delete_embedded("child")

        self.assertEquals(target_doc.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))
        self.assertFalse('child' in doc.embedded)
        self.assertTrue('root' in doc.embedded)

    def testDeleteLastEmbed(self):
        doc = self.make_doc("http://localhost/2")
        doc.embed('root', self.make_doc("http://localhost/"))
        doc.embed('child', self.make_doc("http://localhost/2/1"))

        target_doc = self.make_doc("http://localhost/2")

        doc.delete_embedded('root')
        doc.delete_embedded('child')

        self.assertEquals(target_doc.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))

    def testDeleteIndividualEmbeds(self):
        doc = self.make_doc("http://localhost/2")
        doc.embed('root', self.make_doc("http://localhost/"))
        doc.embed('child', self.make_doc("http://localhost/2/1"))
        doc.embed('child', self.make_doc("http://localhost/2/2"))
        doc.embed('child', self.make_doc("http://localhost/2/3"))

        doc2 = self.make_doc("http://localhost/2")
        doc2.embed('root', self.make_doc("http://localhost/"))
        doc2.embed('child', self.make_doc("http://localhost/2/2"))
        doc2.embed('child', self.make_doc("http://localhost/2/3"))

        doc3 = self.make_doc("http://localhost/2")
        doc3.embed('root', self.make_doc("http://localhost/"))
        doc3.embed('child', self.make_doc("http://localhost/2/2"))

        doc4 = self.make_doc("http://localhost/2")
        doc4.embed('root', self.make_doc("http://localhost/"))

        doc5 = self.make_doc("http://localhost/2")

        doc.delete_embedded("child", "http://localhost/2/1")
        self.assertEquals(doc2.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))

        doc.delete_embedded("child", "http://localhost/2/3")
        self.assertEquals(doc3.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))

        doc.delete_embedded("child", "http://localhost/2/2")
        self.assertEquals(doc4.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))

        doc.delete_embedded("root", "http://localhost/")
        self.assertEquals(doc5.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))

    def testDeleteEmbedsWithoutRel(self):
        doc = self.make_doc("http://localhost/3")
        doc.embed('child', self.make_doc("http://localhost/3/1"))
        doc.embed('child', self.make_doc("http://localhost/3/2"))
        doc.embed('favorite', self.make_doc("http://localhost/3/1"))

        target_doc = self.make_doc("http://localhost/3")
        target_doc.embed('child', self.make_doc("http://localhost/3/2"))

        doc.delete_embedded(href="http://localhost/3/1")

        self.assertEquals(target_doc.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))

    def testDeleteAllEmbeds(self):
        doc = self.make_doc("http://localhost/3")
        doc.embed('child', self.make_doc("http://localhost/3/1"))
        doc.embed('child', self.make_doc("http://localhost/3/2"))
        doc.embed('favorite', self.make_doc("http://localhost/3/1"))

        target_doc = self.make_doc("http://localhost/3")

        doc.delete_embedded()

        self.assertEquals(target_doc.as_object().get(self.EMBEDDED_KEY),
                          doc.as_object().get(self.EMBEDDED_KEY))

    def testDeleteEmbedWithMissingRel(self):
        doc = self.make_doc("http://localhost/3")
        doc.embed('child', self.make_doc("http://localhost/3/1"))
        doc.embed('child', self.make_doc("http://localhost/3/2"))
        doc.embed('favorite', self.make_doc("http://localhost/3/1"))

        target_doc = self.make_doc("http://localhost/3")
        target_doc.embed('child', self.make_doc("http://localhost/3/1"))
        target_doc.embed('child', self.make_doc("http://localhost/3/2"))
        target_doc.embed('favorite', self.make_doc("http://localhost/3/1"))

        missing_rel = ''.join(doc.embedded.keys()) + '_'
        doc.delete_embedded(missing_rel)

        self.assertEquals(target_doc.as_object(), doc.as_object())

    def testDeleteEmbedWithNoEmbeds(self):
        doc = self.make_doc("http://localhost/3")
        target_doc = self.make_doc("http://localhost/3")

        doc.delete_embedded()

        self.assertEquals(target_doc.as_object(), doc.as_object())

    def testAutomaticLink(self):
        doc = self.make_doc("http://localhost/3")
        doc.embed('child', self.make_doc("http://localhost/3/1"))
        doc.delete_embedded('child')

        self.assertAutomaticLink('child' in doc.links)


class DeleteEmbeddedTestsDraft5(DeleteEmbeddedTestsMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5
    assertAutomaticLink = unittest.TestCase.assertTrue


class DeleteEmbeddedTestsDraft4(DeleteEmbeddedTestsMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4
    assertAutomaticLink = unittest.TestCase.assertFalse


class DeleteEmbeddedTestsDraft3(DeleteEmbeddedTestsMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3
    assertAutomaticLink = unittest.TestCase.assertFalse


#


class CurieMutationTestMixin(object):
    def make_doc(self, href, *args, **kwargs):
        result = dougrain.Document.empty("http://localhost", *args,
                                         draft=self.DRAFT, **kwargs)
        result.add_link('self', href)
        return result

    def setUp(self):
        self.doc = self.make_doc("http://localhost/3")
        self.doc.set_curie('rel', "http://localhost/rels/{rel}")

    def testCurieJSONHasCorrectType(self):
        self.assertEquals(type(self.doc.as_object()['_links']['curies']), list)

    def testSetCurie(self):
        new_doc = dougrain.Document(self.doc.as_object(), self.doc.base_uri,
                                    draft=self.DRAFT)
        self.assertEquals("http://localhost/rels/foo",
                          new_doc.expand_curie("rel:foo"))

    def testReplaceCurie(self):
        self.doc.set_curie('rel', "http://localhost/RELS/{rel}.html")

        new_doc = dougrain.Document(self.doc.as_object(), self.doc.base_uri)
        self.assertEquals("http://localhost/RELS/foo.html",
                          new_doc.expand_curie("rel:foo"))

    def testDropCurie(self):
        self.doc.set_curie('tm', "http://www.touchmachine.com/{rel}.html")
        self.doc.drop_curie('rel')

        new_doc = dougrain.Document(self.doc.as_object(), self.doc.base_uri,
                                    draft=self.DRAFT)
        self.assertEquals("rel:foo", self.doc.expand_curie("rel:foo"))
        self.assertEquals("http://www.touchmachine.com/index.html",
                          new_doc.expand_curie("tm:index"))

    def testDropSingleCurie(self):
        self.doc.drop_curie('rel')
        self.assertEquals("rel:foo", self.doc.expand_curie("rel:foo"))


class CurieMutationTestDraft5(CurieMutationTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5

    def testCurieJSONHasCorrectType(self):
        self.assertEquals(type(self.doc.as_object()['_links']['curies']), list)


class CurieMutationTestDraft4(CurieMutationTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5

    def testCurieJSONHasCorrectType(self):
        self.assertEquals(type(self.doc.as_object()['_links']['curies']), list)


class CurieMutationTestDraft3(CurieMutationTestMixin, unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3

    def testCurieJSONHasCorrectType(self):
        self.assertEquals(type(self.doc.as_object()['_links']['curie']), dict)


#


class CurieHidingTests(unittest.TestCase):
    def testCuriesAreNotLinksDraft5(self):
        doc = dougrain.Document({
            '_links': {
                'curies': [{
                    'href': "http://localhost/rel/{rel}",
                    'name': "rel",
                    'templated': True
                }],
                'self': {
                    'href': "http://localhost/0"
                }
            }
        }, "http://localhost/0", draft=dougrain.drafts.DRAFT_5)

        self.assertFalse('curies' in doc.links)

    def testCuriesAreNotLinksDraft4(self):
        doc = dougrain.Document({
            '_links': {
                'curies': [{
                    'href': "http://localhost/rel/{rel}",
                    'name': "rel",
                    'templated': True
                }],
                'self': {
                    'href': "http://localhost/0"
                }
            }
        }, "http://localhost/0", draft=dougrain.drafts.DRAFT_4)

        self.assertFalse('curies' in doc.links)

    def testCuriesAreNotLinksDraft3(self):
        doc = dougrain.Document({
            '_links': {
                'curie': {
                    'href': "http://localhost/rel/{rel}",
                    'name': "rel",
                    'templated': True
                },
                'self': {
                    'href': "http://localhost/0"
                }
            }
        }, "http://localhost/0", draft=dougrain.drafts.DRAFT_3)

        self.assertFalse('curie' in doc.links)

    def testDraft4CuriesAreLinksInDraft43Documents(self):
        modern_doc = dougrain.Document.empty("http://localhost/0",
                                             draft=dougrain.drafts.DRAFT_4)
        modern_doc.set_curie('rel', '/rel/{rel}')
        doc = dougrain.Document.from_object(modern_doc.as_object(),
                                            base_uri="http://localhost/0",
                                            draft=dougrain.drafts.DRAFT_3)
        self.assertTrue('curies' in doc.links)
        self.assertFalse('curie' in doc.links)

    def testDraft5CuriesAreLinksInDraft43Documents(self):
        modern_doc = dougrain.Document.empty("http://localhost/0",
                                             draft=dougrain.drafts.DRAFT_5)
        modern_doc.set_curie('rel', '/rel/{rel}')
        doc = dougrain.Document.from_object(modern_doc.as_object(),
                                            base_uri="http://localhost/0",
                                            draft=dougrain.drafts.DRAFT_3)
        self.assertTrue('curies' in doc.links)
        self.assertFalse('curie' in doc.links)

    def testDraft3CuriesAreLinksInDraft5Document(self):
        doc = dougrain.Document({
            '_links': {
                'curie': {
                    'href': "http://localhost/rel/{rel}",
                    'name': "rel",
                    'templated': True
                },
                'curies': [{
                    'href': "http://localhost/rel/{rel}",
                    'name': "rel",
                    'templated': True
                }],
                'self': {
                    'href': "http://localhost/0"
                }
            }
        }, "http://localhost/0", draft=dougrain.drafts.DRAFT_5)

        self.assertFalse('curies' in doc.links)
        self.assertTrue('curie' in doc.links)

    def testDraft3CuriesAreLinksInDraft4Document(self):
        doc = dougrain.Document({
            '_links': {
                'curie': {
                    'href': "http://localhost/rel/{rel}",
                    'name': "rel",
                    'templated': True
                },
                'curies': [{
                    'href': "http://localhost/rel/{rel}",
                    'name': "rel",
                    'templated': True
                }],
                'self': {
                    'href': "http://localhost/0"
                }
            }
        }, "http://localhost/0", draft=dougrain.drafts.DRAFT_4)

        self.assertFalse('curies' in doc.links)
        self.assertTrue('curie' in doc.links)


#


class LinkCanonicalizationTestsMixin(object):
    def setUp(self):
        self.doc = dougrain.Document.empty("http://localhost/1/",
                                           draft=self.DRAFT)
        self.doc.set_curie("role", "/roles/{rel}")

    def testFindsLinksByFullRelTypeURI(self):
        self.doc.add_link("role:app", "/apps/1")
        self.assertEquals(self.doc.links["role:app"],
                          self.doc.links["http://localhost/roles/app"])

    def testFindsLinksByRelTypePath(self):
        self.doc.add_link("role:app", "/apps/1")
        self.assertEquals(self.doc.links["role:app"],
                          self.doc.links["/roles/app"])

    def testMergesLinksAddedWithSynonymous(self):
        self.doc.add_link("role:app", "/apps/1")
        self.doc.add_link("/roles/app", "/apps/2")
        self.doc.add_link("http://localhost/roles/app", "/apps/3")

        self.assertEquals(["/apps/1", "/apps/2", "/apps/3"],
                          [link.href for link in self.doc.links["role:app"]])

    def testMergesLinkswhenLoading(self):
        links = dict(self.CURIES)
        links.update({
            "self": {"href": "/1"},
            "role:app": {"href": "/apps/1"},
            "/roles/app": {"href": "/apps/2"},
            "http://localhost/roles/app": {"href": "/apps/3"},
        })

        self.doc = dougrain.Document.from_object(
            {"_links": links},
            base_uri="http://localhost/1",
            draft=self.DRAFT)
        self.doc.set_curie("role", "/roles/{rel}")

        self.assertEquals(
            set(["/apps/1", "/apps/2", "/apps/3"]),
            set(link.href for link in self.doc.links["role:app"]))


class LinkCanonicalizationTestsDraft5(LinkCanonicalizationTestsMixin,
                                      unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5
    CURIES = {
        "curies": [
            {"href": "/roles/{rel}", "name": "role", "templated": True}
        ]
    }


class LinkCanonicalizationTestsDraft4(LinkCanonicalizationTestsMixin,
                                      unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4
    CURIES = {
        "curies": [
            {"href": "/roles/{rel}", "name": "role", "templated": True}
        ]
    }


class LinkCanonicalizationTestsDraft3(LinkCanonicalizationTestsMixin,
                                      unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4
    CURIES = {
        "curie": {
            "href": "/roles/{rel}",
            "name": "role",
            "templated": True
        }
    }


#


class EmbeddedCanonicalizationTestMixin(object):
    def setUp(self):
        self.doc = dougrain.Document.empty("http://localhost/1/",
                                           draft=self.DRAFT)
        self.doc.set_curie("role", "/roles/{rel}")

    def new_doc(self, uri):
        doc = dougrain.Document.empty(uri, draft=self.DRAFT)
        doc.add_link('self', uri)
        doc.set_property("name", uri)
        return doc

    def testFindsEmbeddedResourcesByFullRelTypeURI(self):
        self.doc.embed("role:app", self.new_doc("/apps/1"))
        self.assertEquals(self.doc.embedded["role:app"],
                          self.doc.embedded["http://localhost/roles/app"])

    def testFindsEmbeddedResourcesByRelTypePath(self):
        self.doc.embed("role:app", self.new_doc("/apps/1"))
        self.assertEquals(self.doc.embedded["role:app"],
                          self.doc.embedded["/roles/app"])

    def testMergesEmbeddedResourcesAddedWithSynonymous(self):
        self.doc.embed("role:app", self.new_doc("/apps/1"))
        self.doc.embed("/roles/app", self.new_doc("/apps/2"))
        self.doc.embed("http://localhost/roles/app", self.new_doc("/apps/3"))

        self.assertEquals(["/apps/1", "/apps/2", "/apps/3"],
                          [embedded.properties['name']
                           for embedded in self.doc.embedded["role:app"]])


class EmbeddedCanonicalizationTestDraft5(EmbeddedCanonicalizationTestMixin,
                                         unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_5


class EmbeddedCanonicalizationTestDraft4(EmbeddedCanonicalizationTestMixin,
                                         unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_4


class EmbeddedCanonicalizationTestDraft3(EmbeddedCanonicalizationTestMixin,
                                         unittest.TestCase):
    DRAFT = dougrain.drafts.DRAFT_3


#


class EdgeCasesTests(unittest.TestCase):
    def testUrlOfDocumentWithMultipleSelfLinksFromFirstSelfLink(self):
        doc = dougrain.Document.empty("http://localhost")
        doc.add_link('self', "/1")
        doc.add_link('self', "/2")
        self.assertEquals("http://localhost/1", doc.url())

    def testSetReservedAttributeSilentlyFails(self):
        doc = dougrain.Document.empty("http://localhost")
        doc.set_property('_links', {'self': {'href': "/1"}})
        doc.set_property('_embedded', {'child': {'foo': "bar"}})

        self.assertFalse('_links' in doc.properties)
        self.assertIsNone(doc.url())

        self.assertFalse('_embedded' in doc.properties)
        self.assertFalse('child' in doc.embedded)

    def testDeleteReservedAttributeSilentlyFails(self):
        doc = dougrain.Document.empty("http://localhost")
        doc.embed('child', dougrain.Document.from_object({'foo': "bar"},
                                                         "http://localhost"))
        doc.add_link('self', "/1")

        with self.assertRaises(KeyError):
            doc.delete_property('_links')
        with self.assertRaises(KeyError):
            doc.delete_property('_embedded')

        self.assertFalse('_links' in doc.properties)
        self.assertEquals("http://localhost/1", doc.url())

        self.assertFalse('_embedded' in doc.properties)
        self.assertEquals("bar", doc.embedded['child'].properties['foo'])

    def testEmbedDocumentInItself(self):
        doc = dougrain.Document.empty("http://localhost")
        doc.add_link('self', "/1")
        doc.set_curie('rel', "/rels/{rel}")
        doc.set_property('name', "me")
        doc.add_link('rel:other', "/other")
        doc.add_link('next', "/2")

        doc.embed('rel:me', doc)

        self.assertFalse('rel:me' in doc.embedded)


#


class ExplicitDraftTests(unittest.TestCase):
    def testDraft3DocumentHasOldCurieBehaviour(self):
        doc = dougrain.Document.empty("http://localhost",
                                      draft=dougrain.drafts.DRAFT_3)
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_3)
        doc.add_link('self', "/1")
        doc.set_curie('rel', "/rels/{rel}")

        links = doc.as_object()['_links']
        self.assertTrue('curie' in links)
        self.assertFalse('curies' in links)
        curie = links['curie']
        self.assertTrue(isinstance(curie, dict))
        doc.set_curie('foo', "/foos/{rel}")
        curie = links['curie']
        self.assertTrue(isinstance(curie, list))

    def testDraft4DocumentHasNewCurieBehaviour(self):
        doc = dougrain.Document.empty("http://localhost",
                                      draft=dougrain.drafts.DRAFT_4)
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_4)
        doc.add_link('self', "/1")
        doc.set_curie('rel', "/rels/{rel}")

        links = doc.as_object()['_links']
        self.assertFalse('curie' in links)
        self.assertTrue('curies' in links)
        curie = links['curies']
        self.assertTrue(isinstance(curie, list))
        doc.set_curie('foo', "/foos/{rel}")
        curie = links['curies']
        self.assertTrue(isinstance(curie, list))

    def testDraft5DocumentHasNewCurieBehaviour(self):
        doc = dougrain.Document.empty("http://localhost",
                                      draft=dougrain.drafts.DRAFT_5)
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_5)
        doc.add_link('self', "/1")
        doc.set_curie('rel', "/rels/{rel}")

        links = doc.as_object()['_links']
        self.assertFalse('curie' in links)
        self.assertTrue('curies' in links)
        curie = links['curies']
        self.assertTrue(isinstance(curie, list))
        doc.set_curie('foo', "/foos/{rel}")
        curie = links['curies']
        self.assertTrue(isinstance(curie, list))


#


class DraftDetectionTests(unittest.TestCase):
    def testDocumentsWithCurieAreDraft3(self):
        doc = dougrain.Document.from_object(
            {'_links': {'curie': {}}},
            "http://localhost/")
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_3)

    def testDocumentsWithCuriesAreDraft5(self):
        doc = dougrain.Document.from_object(
            {'_links': {'curies': []}},
            "http://localhost/")
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_5)

    def testDocumentsWithNoCurieKeyAreLatest(self):
        doc = dougrain.Document.empty("http://localhost/")
        self.assertEquals(doc.draft, dougrain.drafts.LATEST)

    def testLatestDraftIsDraft5(self):
        doc = dougrain.Document.from_object({}, "http://localhost/",
                                            draft=dougrain.drafts.LATEST)
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_5)

    def testExplicitDraftOverridesAutodetection(self):
        doc = dougrain.Document.from_object(
            {'_links': {'curie': {"href": "/roles/{rel}",
                                  "name": "role",
                                  "templated": True}}},
            "http://localhost/",
            draft=dougrain.drafts.DRAFT_5
        )
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_5)

        doc = dougrain.Document.from_object(
            {'_links': {'curie': {"href": "/roles/{rel}",
                                  "name": "role",
                                  "templated": True}}},
            "http://localhost/",
            draft=dougrain.drafts.DRAFT_4
        )
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_4)

        doc = dougrain.Document.from_object(
            {'_links': {'curies': [{"href": "/roles/{rel}",
                                    "name": "role",
                                    "templated": True}]}},
            "http://localhost/",
            draft=dougrain.drafts.DRAFT_4
        )
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_4)

        doc = dougrain.Document.from_object(
            {'_links': {'curies': []}},
            "http://localhost/",
            draft=dougrain.drafts.DRAFT_3
        )
        self.assertEquals(doc.draft, dougrain.drafts.DRAFT_3)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.
"""
Manipulating HAL documents.
"""

import urlparse
import itertools
import UserDict
from functools import wraps

import curie
import link
from drafts import AUTO
from drafts import LINKS_KEY
from drafts import EMBEDDED_KEY


class CanonicalRels(UserDict.DictMixin, object):
    """Smart querying of link relationship types and link relationships.

    A ``CanonicalRels`` instance is a read-only dictionary-like object that
    provides smart retrieval and de-duplication by link relationship type. It
    is used to make access to links and embedded resources more convenient.

    In addition to well-know link relationship types (eg. ``"self"``), keys can
    be custom link relationship types represented as URIs (eg.
    ``"http://example.com/rels/comments"``), or as URI references (eg.
    ``"/rels/comments"``), or as CURIEs (eg. ``"rel:comments"``).
    ``CanonicalRels`` treats custom link relationship types as equivalent if
    they expand to the same URI, called the canonical key here. Given a
    suitable base URI and set of CURIE templates,
    ``http://example.com/rels/comments``, ``/rels/comments``, and
    ``rel:comments`` are all equivalent.

    ``CanonicalRels`` de-duplicates items with equivalent keys.  De-duplication
    is achieved by appending the new values to the existing values for the
    canonical key. So, ``{"/rels/spam":eggs,"rel:spam":ham}`` becomes
    ``{"http://example.com/rels/spam":[eggs,ham]}``.

    Values can be retrieved using any key that is equivalent to the item's
    canonical key.

    """
    def __init__(self, rels, curies, base_uri, item_filter=lambda _: True):
        """Create a ``CanonicalRels`` instance.

        Arguments:

        - ``rels``:        the relationships to be queried. ``rels`` should be
                           a sequence of ``(key, value)`` tuples or an object
                           with an ``iteritems`` method that returns such a
                           sequence (such as a dictionary). For each tuple in
                           the sequence, ``key`` is a string that identifies
                           the link relationship type and ``value`` is the
                           target of the relationship or a sequence of targets
                           of the relationship.
        - ``curies``:      a ``CurieCollection`` used to expand CURIE keys.
        - ``base_uri``:    URL used as the basis when expanding keys that are
                           relative URI references.
        - ``item_filter``: optional filter on target relationships.
                           ``item_filter`` should be a callable that accepts a
                           target relationship and returns False if the target
                           relationship should be excluded. ``item_filter``
                           will be called once with each target relationship.

        """
        if hasattr(rels, 'iteritems'):
            items = rels.iteritems()
        else:
            items = rels

        self.curies = curies
        self.base_uri = base_uri
        self.rels = {}

        for key, value in items:
            canonical_key = self.canonical_key(key)
            if not canonical_key in self.rels:
                self.rels[canonical_key] = (key, value)
                continue

            original_key, current_value = self.rels[canonical_key]

            new_value = [item for item in current_value if item_filter(item)]
            new_value.extend(item for item in value if item_filter(item))
            self.rels[canonical_key] = original_key, new_value

        self.rels = self.rels

    def canonical_key(self, key):
        """Returns the canonical key for the given ``key``."""
        if key.startswith('/'):
            return urlparse.urljoin(self.base_uri, key)
        else:
            return self.curies.expand(key)

    def original_key(self, key):
        """Returns the first key seen for the given ``key``."""
        return self.rels[self.canonical_key(key)][0]

    def __getitem__(self, key):
        """Returns the link relationship that match the given ``key``.

        ``self[key]`` will return any link relationship who's key is equivalent
        to ``key``. Keys are equivalent if their canonical keys are equal.

        If there is more than one link relationship that matches ``key``, a
        list of matching link relationships is returned.

        If there is one link relationship that matches ``key``, that link
        relationship is returned.

        If there are no link relationships that match ``key``, a ``KeyError``
        is thrown.

        """
        return self.rels[self.canonical_key(key)][1]

    def __contains__(self, key):
        """Returns ``True`` if there are any link relationships for for
        ``self[key].``

        """
        return self.canonical_key(key) in self.rels

    def keys(self):
        """Returns a list of keys that map to every item.

        Each key returned is an original key. That is, the first key
        encountered for the canonical key.

        """
        return [original_key for original_key, _ in self.rels.itervalues()]


class Relationships(UserDict.DictMixin, object):
    """Merged view of relationships from a HAL document.

    Relationships, that is links and embedded resources, are presented as a
    dictionary-like object mapping the full URI of the link relationship type
    to a list of relationships.

    If there are both embedded resources and links for the same link relation
    type, the embedded resources will appear before the links. Otherwise,
    relationships are presented in the order they appear in their respective
    collection.

    Relationships are de-duplicated by their URL, as defined by their ``self``
    link in the case of embedded resources and by their ``href`` in the case of
    links. Only the first relationship with that URL will be included.

    """

    def __init__(self, links, embedded, curies, base_uri):
        """Initialize a ``Relationships`` object.

        Parameters:

        - ``links``:    a dictionary mapping a link relationship type to a
                        ``Link`` instance or a ``list`` of ``Link``
                        instances.
        - ``embedded``: a dictionary mapping a link relationship type to a
                        ``Document`` instance or a ``list`` of ``Document``
                        instances.
        - ``curies``:   a ``CurieCollection`` instance used to expand
                        link relationship type into full link relationship type
                        URLs.

        """
        rels = itertools.chain(embedded.iteritems(), links.iteritems())

        existing_urls = set()

        def item_filter(item):
            url = item.url()
            if url is not None and url in existing_urls:
                return False
            existing_urls.add(item.url())
            return True

        self.canonical_rels = CanonicalRels(rels,
                                            curies,
                                            base_uri,
                                            item_filter)

    def __getitem__(self, key):
        value = self.canonical_rels.__getitem__(key)
        if not isinstance(value, list):
            value = [value]
        return value

    def keys(self):
        return self.canonical_rels.keys()


def mutator(*cache_names):
    """Decorator for ``Document`` methods that change the document.

    This decorator ensures that the object's caches are kept in sync
    when changes are made.

    """
    def deco(fn):
        @wraps(fn)
        def _fn(self, *args, **kwargs):
            try:
                return fn(self, *args, **kwargs)
            finally:
                for cache_name in cache_names:
                    setattr(self, cache_name, None)

        return _fn

    return deco


class Document(object):
    """Represents the document for a HAL resource.

    Constructors:

    - ``Document.empty(base_uri=None)``:
        returns an empty ``Document``.
    - ``Document.from_object(o, base_uri=None, parent_curies=None)``:
        returns a new ``Document`` based on a JSON object.

    Public Instance Attributes:

    - ``properties``: ``dict`` containing the properties of the HAL document,
                      excluding ``_links`` and ``_embedded``. ``properties``
                      should be treated as read-only.
    - ``links``: ``dict`` containing the document's links, excluding
                 ``curies``. Each link relationship type is mapped to a
                 ``Link`` instance or a list of ``Link`` instances. ``links``
                 should be treated as read-only.
    - ``embedded``: dictionary containing the document's embedded resources.
                    Each link relationship type is mapped to a ``Document``
                    instance.
    - ``rels``: a ``Relationships`` instance holding a merged view of the
                relationships from the document.
    - ``draft``: a ``Draft`` instance that selects the version of the spec to
                 which the document should conform. Defaults to
                 ``drafts.AUTO``.

    """
    def __init__(self, o, base_uri, parent_curies=None, draft=AUTO):
        self.prepare_cache()
        self.o = o
        self.base_uri = base_uri
        self.parent_curies = parent_curies
        self.draft = draft.detect(o)

    RESERVED_ATTRIBUTE_NAMES = (LINKS_KEY, EMBEDDED_KEY)

    def properties_cache(self):
        properties = dict(self.o)
        for name in self.RESERVED_ATTRIBUTE_NAMES:
            properties[name] = None
            del properties[name]
        return properties

    def links_cache(self):
        links = {}

        links_json = self.o.get(LINKS_KEY, {})

        for key, value in links_json.iteritems():
            if key == self.draft.curies_rel:
                continue
            links[key] = link.Link.from_object(value, self.base_uri)

        return CanonicalRels(links, self.curies, self.base_uri)

    def curies_cache(self):
        result = curie.CurieCollection()
        if self.parent_curies is not None:
            result.update(self.parent_curies)

        links_json = self.o.get('_links', {})
        curies_json = links_json.get(self.draft.curies_rel)

        if not curies_json:
            return result

        curies = link.Link.from_object(curies_json, self.base_uri)

        if not isinstance(curies, list):
            curies = [curies]

        for curie_link in curies:
            result[curie_link.name] = curie_link

        return result

    def embedded_cache(self):
        embedded = {}
        for key, value in self.o.get(EMBEDDED_KEY, {}).iteritems():
            embedded[key] = self.from_object(value,
                                             self.base_uri,
                                             self.curies)

        return CanonicalRels(embedded, self.curies, self.base_uri)

    def rels_cache(self):
        return Relationships(self.links, self.embedded, self.curies,
                             self.base_uri)

    def prepare_cache(self):
        self._properties_cache = None
        self._curies_cache = None
        self._links_cache = None
        self._embedded_cache = None
        self._rels_cache = None

    @property
    def properties(self):
        if self._properties_cache is None:
            self._properties_cache = self.properties_cache()
        return self._properties_cache

    @property
    def curies(self):
        if self._curies_cache is None:
            self._curies_cache = self.curies_cache()
        return self._curies_cache

    @property
    def links(self):
        if self._links_cache is None:
            self._links_cache = self.links_cache()
        return self._links_cache

    @property
    def embedded(self):
        if self._embedded_cache is None:
            self._embedded_cache = self.embedded_cache()
        return self._embedded_cache

    @property
    def rels(self):
        if self._rels_cache is None:
            self._rels_cache = self.rels_cache()
        return self._rels_cache

    def url(self):
        """Returns the URL for the resource based on the ``self`` link.

        This method returns the ``href`` of the document's ``self`` link if it
        has one, or ``None`` if the document lacks a ``self`` link, or the
        ``href`` of the document's first ``self`` link if it has more than one.

        """
        if not 'self' in self.links:
            return None

        self_link = self.links['self']

        if isinstance(self_link, list):
            for link in self_link:
                return link.url()

        return self_link.url()

    def expand_curie(self, link):
        """Returns the expansion of a CURIE value.

        Arguments:
        - ``link``: a string holding a curie value to expand.

        This method attempts to expand ``link`` using the document's ``curies``
        collection (see ``curie.CurieCollection.expand``).

        """
        return self.curies.expand(link)

    def as_object(self):
        """Returns a dictionary representing the HAL JSON document."""
        return self.o

    def as_link(self):
        """Returns a ``Link`` to the resource."""
        return self.links['self']

    @mutator('_properties_cache')
    def set_property(self, key, value):
        """Set a property on the document.

        Calling code should use this method to add and modify properties
        on the document instead of modifying ``properties`` directly.

        If ``key`` is ``"_links"`` or ``"_embedded"`` this method will silently
        fail.

        If there is no property with the name in ``key``, a new property is
        created with the name from ``key`` and the value from ``value``. If
        the document already has a property with that name, it's value
        is replaced with the value in ``value``.

        """
        if key in self.RESERVED_ATTRIBUTE_NAMES:
            return
        self.o[key] = value

    @mutator('_properties_cache')
    def delete_property(self, key):
        """Remove a property from the document.

        Calling code should use this method to remove properties on the
        document instead of modifying ``properties`` directly.

        If there is a property with the name in ``key``, it will be removed.
        Otherwise, a ``KeyError`` will be thrown.

        """
        if key in self.RESERVED_ATTRIBUTE_NAMES:
            raise KeyError(key)
        del self.o[key]

    def link(self, href, **kwargs):
        """Retuns a new link relative to this resource."""
        return link.Link(dict(href=href, **kwargs), self.base_uri)

    @mutator('_links_cache')
    def add_link(self, rel, target, wrap=False, **kwargs):
        """Adds a link to the document.

        Calling code should use this method to add links instead of
        modifying ``links`` directly.

        This method adds a link to the given ``target`` to the document with
        the given ``rel``. If one or more links are already present for that
        link relationship type, the new link will be added to the existing
        links for that link relationship type.

        If ``target`` is a string, a link is added with ``target`` as its
        ``href`` property and other properties from the keyword arguments.

        If ``target`` is a ``Link`` object, it is added to the document and the
        keyword arguments are ignored.

        If ``target`` is a ``Document`` object, ``target``'s ``self`` link is
        added to this document and the keyword arguments are ignored.

        If ``target`` is a ``Builder`` object, ``target``'s ``self`` link is
        added to this document and the keyword arguments are ignored.

        Arguments:

        - ``rel``: a string specifying the link relationship type of the link.
          It should be a well-known link relation name from the IANA registry
          (http://www.iana.org/assignments/link-relations/link-relations.xml),
          a full URI, or a CURIE.
        - ``target``: the destination of the link.
        - ``wrap``: Defaults to False, but if True, specifies that the link
          object should be initally wrapped in a JSON array even if it is the
          first link for the given ``rel``.

        """
        if hasattr(target, 'as_link'):
            link = target.as_link()
        else:
            link = self.link(target, **kwargs)

        links = self.o.setdefault(LINKS_KEY, {})

        new_link = link.as_object()
        collected_links = CanonicalRels(links, self.curies, self.base_uri)
        if rel not in collected_links:
            if wrap:
                links[rel] = [new_link]
            else:
                links[rel] = new_link
            return

        original_rel = collected_links.original_key(rel)

        current_links = links[original_rel]
        if isinstance(current_links, list):
            current_links.append(new_link)
        else:
            links[original_rel] = [current_links, new_link]

    @mutator('_links_cache')
    def delete_link(self, rel=None, href=lambda _: True):
        """Deletes links from the document.

        Calling code should use this method to remove links instead of
        modyfying ``links`` directly.

        The optional arguments, ``rel`` and ``href`` are used to select the
        links that will be deleted. If neither of the optional arguments are
        given, this method deletes every link in the document. If ``rel`` is
        given, only links for the matching link relationship type are deleted.
        If ``href`` is given, only links with a matching ``href`` are deleted.
        If both ``rel`` and ``href`` are given, only links with matching
        ``href`` for the matching link relationship type are delted.

        Arguments:

        - ``rel``: an optional string specifying the link relationship type of
                   the links to be deleted.
        - ``href``: optionally, a string specifying the ``href`` of the links
                    to be deleted, or a callable that returns true when its
                    single argument is in the set of ``href``s to be deleted.

        """
        if not LINKS_KEY in self.o:
            return

        links = self.o[LINKS_KEY]
        if rel is None:
            for rel in links.keys():
                self.delete_link(rel, href)
            return

        if callable(href):
            href_filter = href
        else:
            href_filter = lambda x: x == href

        links_for_rel = links.setdefault(rel, [])
        if isinstance(links_for_rel, dict):
            links_for_rel = [links_for_rel]

        new_links_for_rel = []
        for link in links_for_rel:
            if not href_filter(link['href']):
                new_links_for_rel.append(link)

        if new_links_for_rel:
            if len(new_links_for_rel) == 1:
                new_links_for_rel = new_links_for_rel[0]

            links[rel] = new_links_for_rel
        else:
            del links[rel]

        if not self.o[LINKS_KEY]:
            del self.o[LINKS_KEY]

    @classmethod
    def from_object(cls, o, base_uri=None, parent_curies=None, draft=AUTO):
        """Returns a new ``Document`` based on a JSON object or array.

        Arguments:

        - ``o``: a dictionary holding the deserializated JSON for the new
                 ``Document``, or a ``list`` of such documents.
        - ``base_uri``: optional URL used as the basis when expanding
                               relative URLs in the document.
        - ``parent_curies``: optional ``CurieCollection`` instance holding the
                             CURIEs of the parent document in which the new
                             document is to be embedded. Calling code should
                             not normally provide this argument.
        - ``draft``: a ``Draft`` instance that selects the version of the spec
                     to which the document should conform. Defaults to
                     ``drafts.AUTO``.

        """

        if isinstance(o, list):
            return map(lambda x: cls.from_object(x,
                                                 base_uri,
                                                 parent_curies,
                                                 draft),
                       o)

        return cls(o, base_uri, parent_curies, draft)

    @classmethod
    def empty(cls, base_uri=None, draft=AUTO):
        """Returns an empty ``Document``.

        Arguments:

        - ``base_uri``: optional URL used as the basis when expanding
                               relative URLs in the document.
        - ``draft``: a ``Draft`` instance that selects the version of the spec
                     to which the document should conform. Defaults to
                     ``drafts.AUTO``.
        """
        return cls.from_object({}, base_uri=base_uri, draft=draft)

    @mutator('_embedded_cache')
    def embed(self, rel, other, wrap=False):
        """Embeds a document inside this document.

        Arguments:

        - ``rel``: a string specifying the link relationship type of the
          embedded resource. ``rel`` should be a well-known link relation name
          from the IANA registry
          (http://www.iana.org/assignments/link-relations/link-relations.xml),
          a full URI, or a CURIE.
        - ``other``: a ``Document`` instance that will be embedded in this
          document. If ``other`` is identical to this document, this method
          will silently fail.
        - ``wrap``: Defaults to False, but if True, specifies that the embedded
          resource object should be initally wrapped in a JSON array even if it
          is the first embedded resource for the given ``rel``.

        Calling code should use this method to add embedded resources instead
        of modifying ``embedded`` directly.

        This method embeds the given document in this document with the given
        ``rel``. If one or more documents have already been embedded for that
        ``rel``, the new document will be embedded in addition to those
        documents.

        """

        if other == self:
            return

        embedded = self.o.setdefault(EMBEDDED_KEY, {})
        collected_embedded = CanonicalRels(embedded,
                                           self.curies,
                                           self.base_uri)

        if rel not in collected_embedded:
            if wrap:
                embedded[rel] = [other.as_object()]
            else:
                embedded[rel] = other.as_object()
        else:
            original_rel = collected_embedded.original_key(rel)

            current_embedded = embedded[original_rel]
            if isinstance(current_embedded, list):
                current_embedded.append(other.as_object())
            else:
                embedded[original_rel] = [current_embedded, other.as_object()]

        if not self.draft.automatic_link:
            return

        url = other.url()
        if not url:
            return

        if url in (link.url() for link in self.links.get(rel, [])):
            return

        self.add_link(rel, other, wrap=wrap)

    @mutator('_embedded_cache')
    def delete_embedded(self, rel=None, href=lambda _: True):
        """Removes an embedded resource from this document.

        Calling code should use this method to remove embedded resources
        instead of modifying ``embedded`` directly.

        The optional arguments, ``rel`` and ``href`` are used to select the
        embedded resources that will be removed. If neither of the optional
        arguments are given, this method removes every embedded resource from
        this document. If ``rel`` is given, only embedded resources for the
        matching link relationship type are removed. If ``href`` is given, only
        embedded resources with a ``self`` link matching ``href`` are deleted.
        If both ``rel`` and ``href`` are given, only embedded resources with
        matching ``self`` link for the matching link relationship type are
        removed.

        Arguments:

        - ``rel``: an optional string specifying the link relationship type of
                   the embedded resources to be removed.
        - ``href``: optionally, a string specifying the ``href`` of the
                    ``self`` links of the resources to be removed, or a
                    callable that returns true when its single argument matches
                    the ``href`` of the ``self`` link of one of the resources
                    to be removed.

        """
        if EMBEDDED_KEY not in self.o:
            return

        if rel is None:
            for rel in self.o[EMBEDDED_KEY].keys():
                self.delete_embedded(rel, href)
            return

        if rel not in self.o[EMBEDDED_KEY]:
            return

        if callable(href):
            url_filter = href
        else:
            url_filter = lambda x: x == href

        rel_embeds = self.o[EMBEDDED_KEY][rel]

        if isinstance(rel_embeds, dict):
            del self.o[EMBEDDED_KEY][rel]

            if not self.o[EMBEDDED_KEY]:
                del self.o[EMBEDDED_KEY]
            return

        new_rel_embeds = []
        for embedded in list(rel_embeds):
            embedded_doc = Document(embedded, self.base_uri)
            if not url_filter(embedded_doc.url()):
                new_rel_embeds.append(embedded)

        if not new_rel_embeds:
            del self.o[EMBEDDED_KEY][rel]
        elif len(new_rel_embeds) == 1:
            self.o[EMBEDDED_KEY][rel] = new_rel_embeds[0]
        else:
            self.o[EMBEDDED_KEY][rel] = new_rel_embeds

        if not self.o[EMBEDDED_KEY]:
            del self.o[EMBEDDED_KEY]

    @mutator('_curies_cache')
    def set_curie(self, name, href):
        """Sets a CURIE.

        A CURIE link with the given ``name`` and ``href`` is added to the
        document.

        """

        self.draft.set_curie(self, name, href)

    @mutator('_curies_cache')
    def drop_curie(self, name):
        """Removes a CURIE.

        The CURIE link with the given name is removed from the document.

        """
        curies = self.o[LINKS_KEY][self.draft.curies_rel]
        if isinstance(curies, dict) and curies['name'] == name:
            del self.o[LINKS_KEY][self.draft.curies_rel]
            return

        for i, curie in enumerate(curies):
            if curie['name'] == name:
                del curies[i]
                break

            continue

    def __iter__(self):
        yield self

    def __eq__(self, other):
        if not isinstance(other, Document):
            return False

        return self.as_object() == other.as_object()

    def __repr__(self):
        return "<Document %r>" % self.url()



# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.
"""
Creating HAL documents.
"""

from dougrain import drafts
from dougrain import link


class Builder(object):
    """Simplify creation of HAL documents.

    ``Builder`` provides a lightweight chainable API for creating HAL
    documents.

    Unlike ``dougrain.Document``, ``Builder`` provides no facilities for
    interrogating or mutating existing documents and makes. ``Builder`` also
    makes fewer sanity checks than ``dougrain.Document``, which makes it
    considerably faster, but makes it more likely to produce invalid HAL
    documents.

    """

    def __init__(self, href, draft=drafts.LATEST, **kwargs):
        """``Builder(href, draft=drafts.LATEST, **kwargs)``

        Make a builder for a document. The document starts with a ``self``
        link to ``href``.

        The version of the spec may be specified in the optional ``draft``
        argument, which defaults to the latest draft. For example, to build a
        document conforming to the older Draft 4 of the spec:

        ``
        dougrain.Document("/test", draft=dougrain.drafts.DRAFT_4)
        ``

        Additional properties for the self link may be supplied in other
        keyword arguments. For example:

        ``
        dougrain.Document("/test", profile="/profiles/")
        ``

        """
        self.o = {'_links': {'self': dict(href=href, **kwargs)}}
        self.draft = draft.draft

    def url(self):
        """Returns the URL for the resource based on the ``self`` link.

        This method is used when embedding a ``Builder`` in a
        ``dougrain.Document``.
        """
        return self.o['_links']['self']['href']

    def as_object(self):
        """Returns a dictionary representing the HAL JSON document."""
        return self.o

    def as_link(self):
        """Returns a ``Link`` to the document.

        This method is used when adding a link to a ``Builder`` from a
        ``dougrain.Document``.
        """
        return link.Link(self.o['_links']['self'], None)

    def add_curie(self, name, href):
        """Adds a CURIE definition.

        A CURIE link with the given ``name`` and ``href`` is added to the
        document.

        This method returns self, allowing it to be chained with additional
        method calls.
        """
        self.draft.set_curie(self, name, href)
        return self

    def set_property(self, name, value):
        """Set a property on the document.

        If there is no property with the name in ``name``, a new property is
        created with the name from ``name`` and the value from ``value``. If
        the document already has a property with that name, it's value
        is replaced with the value in ``value``.

        This method returns self, allowing it to be chained with additional
        method calls.

        WARNING: ``name`` should not be one of the reserved property
        names (``'_links'`` or ``'_embedded'``). If ``name`` is ``'_links'`` or
        ``'_embedded'``, this method may silently corrupt the JSON object
        representation and cause undefined behaviour later.

        """
        self.o[name] = value
        return self

    def add_link(self, rel, target, wrap=False, **kwargs):
        """Adds a link to the document.

        This method adds a link to the given ``target`` to the document with
        the given ``rel``. If one or more links are already present for that
        link relationship type, the new link will be added to the existing
        links for that link relationship type.

        Unlike ``dougrain.Document.add_link``, this method does not detect
        equivalence between relationship types with different representations.

        If ``target`` is a string, a link is added with ``target`` as its
        ``href`` property and other properties from the keyword arguments.

        If ``target`` is a ``dougrain.Document`` object, a link is added with
        ``target``'s URL as its ``href`` property and other property from the
        keyword arguments.

        If ``target`` is a ``Builder`` object, a link is added with
        ``target``'s URL as its ``href`` property and other property from the
        keyword arguments.

        This method returns self, allowing it to be chained with additional
        method calls.

        Arguments:

        - ``rel``: a string specifying the link relationship type of the link.
          It should be a well-known link relation name from the IANA registry
          (http://www.iana.org/assignments/link-relations/link-relations.xml),
          a full URI, or a CURIE.
        - ``target``: the destination of the link.
        - ``wrap``: Defaults to False, but if True, specifies that the link
          object should be initally wrapped in a JSON array even if it is the
          first link for the given ``rel``.

        """
        if isinstance(target, str) or isinstance(target, unicode):
            new_link = dict(href=target, **kwargs)
        else:
            new_link = dict(href=target.url(), **kwargs)

        self._add_rel('_links', rel, new_link, wrap)
        return self

    def embed(self, rel, target, wrap=False):
        """Embeds a document inside this document.

        This method embeds the given document in this document with the given
        ``rel``. If one or more documents have already been embedded for that
        ``rel``, the new document will be embedded in addition to those
        documents.

        This method returns self, allowing it to be chained with additional
        method calls.

        Arguments:

        - ``rel``: a string specifying the link relationship type of the
          embedded resource. ``rel`` should be a well-known link relation name
          from the IANA registry
          (http://www.iana.org/assignments/link-relations/link-relations.xml),
          a full URI, or a CURIE.
        - ``target``: a ``Builder`` instance or a ``dougrain.Document``
          instance that will be embedded in this document.
        - ``wrap``: Defaults to False, but if True, specifies that the embedded
          resource object should be initally wrapped in a JSON array even if it
          is the first embedded resource for the given ``rel``.

        Unlike ``dougrain.Document.embed``, this method does not detect
        equivalence between relationship types with different representations.

        WARNING: ``target`` should not be identical to ``self`` or any document
        that embeds ``self``.
        """
        new_embed = target.as_object()
        self._add_rel('_embedded', rel, new_embed, wrap)

        if self.draft.automatic_link:
            self.add_link(rel, target, wrap)

        return self

    def _add_rel(self, key, rel, thing, wrap):
        """Adds ``thing`` to links or embedded resources.

        Calling code should not use this method directly and should use
        ``embed`` or ``add_link`` instead.

        """
        self.o.setdefault(key, {})

        if wrap:
            self.o[key].setdefault(rel, [])

        if rel not in self.o[key]:
            self.o[key][rel] = thing
            return

        existing = self.o[key].get(rel)
        if isinstance(existing, list):
            existing.append(thing)
            return

        self.o[key][rel] = [existing, thing]



# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.
"""Draft-specific behaviours.

This module represents the differences between drafts of the HAL specification
and provides a mechanism for specifying the backwards-compatibility behavior of
Document instances.

Calling code is expected to use the following members:

    - ``DRAFT_3`` : the document should be compatible with Draft 3.
    - ``DRAFT_4`` : the document should be compatible with Draft 4.
    - ``DRAFT_5`` : the document should be compatible with Draft 5.
    - ``LATEST``  : the document should be compatible with the latest draft
                    implemented in this library.
    - ``AUTO``    : the document should be compatible with a draft level it
                    guesses from the JSON object with which it is initialized.

"""

LINKS_KEY = '_links'
EMBEDDED_KEY = '_embedded'


class Draft3(object):
    """Behaviour that is compatibile with draft 3 of the HAL spec.

    CURIEs are stored in a link with the relation type 'curie'. If there are
    multiple CURIEs, the document's JSON representation has a JSON array in
    ``_links.curie``, but if there is only one CURIE, ``_links.curie`` is a
    JSON object.

    See http://tools.ietf.org/html/draft-kelly-json-hal-03.
    """

    curies_rel = 'curie'
    automatic_link = False

    def set_curie(self, doc, name, href):
        doc.add_link(self.curies_rel, href, name=name, templated=True)


class Draft4(Draft3):
    """Behaviour that is compatibile with draft 4 of the HAL spec.

    CURIEs are stored in a link with the relation type 'curies'. If there
    are one or more CURIEs, the document's JSON representation has a JSON
    array in ``_links.curies``.

    See http://tools.ietf.org/html/draft-kelly-json-hal-04.
    """

    curies_rel = 'curies'

    def set_curie(self, doc, name, href):
        # CURIE links should always be in an array, even if there is only
        # one.
        doc.o.setdefault(LINKS_KEY, {}).setdefault(self.curies_rel, [])
        doc.o[LINKS_KEY][self.curies_rel].append(
            {'href': href, 'name': name, 'templated': True})


class Draft5(Draft4):
    """Behaviour that is compatibile with draft 5 of the HAL spec.

    CURIEs are stored in a link with the relation type 'curies'. If there
    are one or more CURIEs, the document's JSON representation has a JSON
    array in ``_links.curies``.

    If a document with a 'self' link is embedded and there is no
    corresponding link, a corresponding link is added.

    See http://tools.ietf.org/html/draft-kelly-json-hal-05.
    """

    automatic_link = True


class DraftIdentifier(object):
    """Identifies HAL draft level of a JSON document.

    When created with an existing JSON object, the document guesses the
    draft version based on the presence of a link with a relation type of
    'curie' or 'curies'.
    """

    def detect(self, obj):
        """Identifies the HAL draft level of a given JSON object."""
        links = obj.get(LINKS_KEY, {})

        for detector in [LATEST, DRAFT_3]:
            if detector.draft.curies_rel in links:
                return detector.detect(obj)

        return LATEST.detect(obj)

    def __repr__(self):
        return "%s()" % self.__class__.__name__


class FixedDraftIdentifier(object):
    """Identifies JSON objects as having a known draft level."""

    def __init__(self, draft):
        self.draft = draft

    def detect(self, obj):
        """Identify the HAL draft level of obj as this instance's draft."""
        return self.draft

    def __eq__(self, other):
        if isinstance(other, FixedDraftIdentifier):
            return other.draft == self.draft

        return other == self.draft

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__,
                           self.draft.__class__.__name__)


DRAFT_3 = FixedDraftIdentifier(Draft3())
DRAFT_4 = FixedDraftIdentifier(Draft4())
DRAFT_5 = FixedDraftIdentifier(Draft5())
LATEST = DRAFT_5
AUTO = DraftIdentifier()



#!/usr/bin/python
# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.

import unittest
from dougrain import Builder, Document, drafts


class BuilderTests(unittest.TestCase):
    def setUp(self):
        self.uri = "http://localhost/%s/%s" % (self.__class__.__name__,
                                               self._testMethodName)
        self.builder = Builder(self.uri)


class InitialBuilderTests(BuilderTests):
    def testStartsWithNoProperties(self):
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertEqual(doc.properties.keys(), [])

    def testAlreadyHasSelfLink(self):
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIn('self', doc.links)

        self_link = doc.links['self']
        self.assertEqual(self_link.url(), self.uri)
        self.assertEqual(doc.url(), self.uri)
        self.assertEqual(doc.links.keys(), ['self'])

    def testStartsWithNoEmbeds(self):
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertEqual(doc.embedded.keys(), [])

    def testAddsKeywordArgumentsToSelfLink(self):
        self.builder = Builder(self.uri,
                               name="test",
                               title="Test Document",
                               type="application/hal+json",
                               profile="/profiles/root",
                               hreflang="en-GB",
                               deprecation=False,
                               templated=False)
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhsot")
        self_link = doc.links['self']
        self.assertEqual(self_link.href, self.uri)
        self.assertEqual(self_link.name, "test")
        self.assertEqual(self_link.title, "Test Document")
        self.assertEqual(self_link.type, "application/hal+json")
        self.assertEqual(self_link.profile, "/profiles/root")
        self.assertFalse(self_link.deprecation)
        self.assertFalse(self_link.is_templated)


class PropertyBuilderTests(BuilderTests):
    def testSetNewProperty(self):
        self.builder.set_property('spam', "foo")

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")

        self.assertEquals(doc.properties['spam'], "foo")


class LinkBuilderTestsMixin(object):
    def make_target(self, href):
        raise NotImplemented

    def add_link(self, rel, href, **kwargs):
        self.builder.add_link(rel, self.make_target(href), **kwargs)

    def testAddSimpleLink(self):
        self.add_link('item', "/items/1")

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIn('item', doc.links)

        item_link = doc.links['item']
        self.assertEqual(item_link.url(), doc.link("/items/1").url())

    def testAddLinkWithOptionalProperties(self):
        self.add_link('item', "/items/{item_id}",
                      name='name',
                      title='Title',
                      type='application/hal+json',
                      profile='/profiles/item',
                      hreflang='en',
                      deprecation='/deprecation/item',
                      templated=True)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_link = doc.links['item']
        self.assertEqual(item_link.url(item_id='1'),
                         "http://localhost/items/1")
        self.assertEqual(item_link.name, 'name')
        self.assertEqual(item_link.title, 'Title')
        self.assertEqual(item_link.type, 'application/hal+json')
        self.assertEqual(item_link.profile, '/profiles/item')
        self.assertEqual(item_link.hreflang, 'en')
        self.assertEqual(item_link.deprecation, '/deprecation/item')
        self.assertTrue(item_link.is_templated)

    def testAddSecondLink(self):
        self.add_link('item', '/items/1', name='first')
        self.add_link('item', '/items/2', name='second')

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_links = doc.links['item']
        self.assertSequenceEqual([item_link.name for item_link in item_links],
                                 ['first', 'second'])

    def testAddThirdLink(self):
        self.add_link('item', '/items/1', name='first')
        self.add_link('item', '/items/2', name='second')
        self.add_link('item', '/items/3', name='third')

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_links = doc.links['item']
        self.assertSequenceEqual([item_link.name for item_link in item_links],
                                 ['first', 'second', 'third'])

    def testAddWrappedLink(self):
        self.add_link('item', '/items/1', wrap=True)
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIn('item', doc.links)

        item_link = doc.links['item']
        self.assertEqual(item_link.url(), doc.link("/items/1").url())

    def testAddSecondWrappedLink(self):
        self.add_link('item', '/items/1', name='first', wrap=True)
        self.add_link('item', '/items/2', name='second', wrap=True)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_links = doc.links['item']
        self.assertSequenceEqual([item_link.name for item_link in item_links],
                                 ['first', 'second'])

    def testFirstUnwrappedLinkIsDict(self):
        self.add_link('item', '/items/1', wrap=False)
        self.assertIsInstance(self.builder.as_object()['_links']['item'],
                              dict)

    def testFirstWrappedLinkIsList(self):
        self.add_link('item', '/items/1', wrap=True)
        self.assertIsInstance(self.builder.as_object()['_links']['item'],
                              list)

    def testDefaultLinkIsNotWrapped(self):
        self.add_link('item', '/items/1')
        self.assertIsInstance(self.builder.as_object()['_links']['item'],
                              dict)

    def testStartsWithNoCuries(self):
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertSequenceEqual(doc.curies.keys(), [])

        self.assertEqual(doc.expand_curie('rel:spam'), "rel:spam")


class HrefLinkBuilderTests(BuilderTests, LinkBuilderTestsMixin):
    def make_target(self, href):
        return href


class UnicodeHrefLinkBuilderTests(BuilderTests, LinkBuilderTestsMixin):
    def make_target(self, href):
        return unicode(href)


class BuilderLinkBuilderTests(BuilderTests, LinkBuilderTestsMixin):
    def make_target(self, href):
        return Builder(href)


class DocumentLinkBuilderTests(BuilderTests, LinkBuilderTestsMixin):
    def make_target(self, href):
        doc = Document.empty("http://localhost/")
        doc.add_link('self', href)
        return doc


class CurieBuilderTests(BuilderTests):
    def testAddCurie(self):
        self.builder.add_curie('rel', "/rels/{rel}")
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertEqual(doc.expand_curie('rel:spam'),
                         "http://localhost/rels/spam")

    def testAddSecondCurie(self):
        self.builder.add_curie('rel', "/rels/{rel}")
        self.builder.add_curie('admin', "/rels/admin/{rel}")
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertEqual(doc.expand_curie('rel:spam'),
                         "http://localhost/rels/spam")
        self.assertEqual(doc.expand_curie('admin:eggs'),
                         "http://localhost/rels/admin/eggs")

    def testAddCurieDraft5(self):
        self.builder = Builder(self.uri, draft=drafts.DRAFT_5)
        self.builder.add_curie('rel', "/rels/{rel}")
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost",
                                   draft=drafts.DRAFT_5)
        self.assertEqual(doc.expand_curie('rel:spam'),
                         "http://localhost/rels/spam")

        curies = self.builder.as_object()['_links']['curies']
        self.assertIsInstance(curies, list)

    def testAddCurieDraft4(self):
        self.builder = Builder(self.uri, draft=drafts.DRAFT_4)
        self.builder.add_curie('rel', "/rels/{rel}")
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost",
                                   draft=drafts.DRAFT_4)
        self.assertEqual(doc.expand_curie('rel:spam'),
                         "http://localhost/rels/spam")

        curies = self.builder.as_object()['_links']['curies']
        self.assertIsInstance(curies, list)

    def testAddCurieDraft3(self):
        self.builder = Builder(self.uri, draft=drafts.DRAFT_3)
        self.builder.add_curie('rel', "/rels/{rel}")
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost",
                                   draft=drafts.DRAFT_3)
        self.assertEqual(doc.expand_curie('rel:spam'),
                         "http://localhost/rels/spam")

        curies = self.builder.as_object()['_links']['curie']
        self.assertIsInstance(curies, dict)


class EmbedBuilderTests(BuilderTests):
    def make_target(self, name):
        target_uri = "http://localhost/%s/%s/%s" % (self.__class__.__name__,
                                                    self._testMethodName,
                                                    name)
        target_builder = Builder(target_uri)
        target_builder.set_property('name', name)
        return target_builder

    def testFirstEmbedDefault(self):
        target = self.make_target("first")
        self.builder.embed('item', target)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIsInstance(self.builder.as_object()['_embedded']['item'],
                              dict)
        self.assertEquals(doc.embedded['item'].properties['name'], "first")

    def testFirstEmbedNotWrapped(self):
        target = self.make_target("first")
        self.builder.embed('item', target, wrap=False)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIsInstance(self.builder.as_object()['_embedded']['item'],
                              dict)
        self.assertEquals(doc.embedded['item'].properties['name'], "first")

    def testFirstEmbedWrapped(self):
        target = self.make_target("first")
        self.builder.embed('item', target, wrap=True)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIsInstance(self.builder.as_object()['_embedded']['item'],
                              list)
        item_embeds = doc.embedded['item']
        names = [embedded.properties['name'] for embedded in item_embeds]
        self.assertEquals(names, ["first"])

    def testSecondEmbed(self):
        self.builder.embed('item', self.make_target('first'))
        self.builder.embed('item', self.make_target('second'))

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_embeds = doc.embedded['item']
        names = [embedded.properties['name'] for embedded in item_embeds]
        self.assertSequenceEqual(names, ['first', 'second'])

    def testThirdEmbed(self):
        self.builder.embed('item', self.make_target('first'))
        self.builder.embed('item', self.make_target('second'))
        self.builder.embed('item', self.make_target('third'))

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_embeds = doc.embedded['item']
        names = [embedded.properties['name'] for embedded in item_embeds]
        self.assertSequenceEqual(names, ['first', 'second', 'third'])

    def testThreeWrappedEmbeds(self):
        self.builder.embed('item', self.make_target('first'), wrap=True)
        self.builder.embed('item', self.make_target('second'), wrap=True)
        self.builder.embed('item', self.make_target('third'), wrap=True)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_embeds = doc.embedded['item']
        names = [embedded.properties['name'] for embedded in item_embeds]
        self.assertSequenceEqual(names, ['first', 'second', 'third'])

    def testEmbedAlsoAddsLinkWithDraft5(self):
        self.builder = Builder(self.uri, draft=drafts.DRAFT_5)
        self.builder.embed('item', self.make_target('first'))
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_links = doc.links['item']
        urls = [link.url() for link in item_links]
        item_embeds = doc.embedded['item']
        expected_urls = [embedded.url() for embedded in item_embeds]
        self.assertSequenceEqual(urls, expected_urls)

    def testEmbedDoesNotAddLinkWithDraft4(self):
        self.builder = Builder(self.uri, draft=drafts.DRAFT_4)
        self.builder.embed('item', self.make_target('first'))
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_links = doc.links.get('item', [])
        urls = [link.url() for link in item_links]
        item_embeds = doc.embedded['item']
        expected_urls = []
        self.assertSequenceEqual(urls, expected_urls)

    def testEmbedDoesNotAddLinkWithDraft3(self):
        self.builder = Builder(self.uri, draft=drafts.DRAFT_3)
        self.builder.embed('item', self.make_target('first'))
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        item_links = doc.links.get('item', [])
        urls = [link.url() for link in item_links]
        item_embeds = doc.embedded['item']
        expected_urls = []
        self.assertSequenceEqual(urls, expected_urls)

    def testEmbedAddsUnnecessaryLinkWithDraft5(self):
        self.builder = Builder(self.uri, draft=drafts.DRAFT_5)
        target = self.make_target('first')

        self.builder.add_link('item', target, wrap=True)
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        urls_before_embed = [link.url() for link in doc.links['item']]

        self.builder.embed('item', target, wrap=True)
        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        urls_after_embed = [link.url() for link in doc.links['item']]

        item_embeds = doc.embedded['item']
        embedded_urls = [embedded.url() for embedded in item_embeds]
        expected_urls = urls_before_embed + embedded_urls
        self.assertSequenceEqual(urls_after_embed, expected_urls)

    def testAutomaticLinkIsNotWrappedDefault(self):
        target = self.make_target("first")
        self.builder.embed('item', target)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIsInstance(self.builder.as_object()['_links']['item'],
                              dict)

    def testAutomaticLinkIsNotWrappedWithUnwrappedEmbed(self):
        target = self.make_target("first")
        self.builder.embed('item', target, wrap=False)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIsInstance(self.builder.as_object()['_links']['item'],
                              dict)

    def testAutomaticLinkIsWrappedWithWrappedEmbed(self):
        target = self.make_target("first")
        self.builder.embed('item', target, wrap=True)

        doc = Document.from_object(self.builder.as_object(),
                                   base_uri="http://localhost")
        self.assertIsInstance(self.builder.as_object()['_links']['item'],
                              list)


class EmbedBuilderDocumentTests(EmbedBuilderTests):
    def make_target(self, name):
        target_uri = "http://localhost/%s/%s/%s" % (self.__class__.__name__,
                                                    self._testMethodName,
                                                    name)
        target_doc = Document.empty("http://localhost")
        target_doc.add_link('self', target_uri)
        target_doc.set_property('name', name)
        return target_doc


class ChainingBuilderTests(unittest.TestCase):
    def testChainAfterSetProperty(self):
        obj = (Builder("/item/1")
               .set_property("name", "Thing #1")
               .set_property("chained", True)).as_object()

        doc = Document.from_object(obj, base_uri="http://localhost/")

        self.assertTrue(doc.properties['chained'])

    def testChainAfterAddLink(self):
        obj = (Builder("/item/1")
               .add_link("next", "/item/2")
               .set_property("chained", True)).as_object()

        doc = Document.from_object(obj, base_uri="http://localhost/")

        self.assertTrue(doc.properties['chained'])

    def testChainAfterAddCurie(self):
        obj = (Builder("/item/1")
               .add_curie("rel", "/rels/{rel}")
               .set_property("chained", True)).as_object()

        doc = Document.from_object(obj, base_uri="http://localhost/")

        self.assertTrue(doc.properties['chained'])

    def testChainAfterEmbed(self):
        target = Builder("/item/2")
        obj = (Builder("/item/1")
               .embed("next", target)
               .set_property("chained", True)).as_object()

        doc = Document.from_object(obj, base_uri="http://localhost/")

        self.assertTrue(doc.properties['chained'])

    def testBuildDocument(self):
        obj = (Builder("/products/254-rocket-skates")
               .add_curie('api', '/rels/{rel}')
               .set_property('name', "Rocket Skates")
               .set_property(
                   'description',
                   "Endorsed by top athletes, these affordable ...")
               .set_property('ean', "55-5555-5555-X")
               .add_link('api:reviews', "/products/254-rocket-skates/reviews")
               .embed('api:pictures',
                      Builder("/products/254-rocket-skates/pictures/")
                      .set_property('count', 2)
                      .add_link('item',
                                "/products/254-rocket-skates/pictures/1")
                      .add_link('item',
                                "/products/254-rocket-skates/pictures/2")
                      .embed('item',
                             Builder("/products/254-rocket-skates/pictures/1")
                             .set_property('count', 1)
                             .add_link('api:subject',
                                       "/products/254-rocket-skates")
                             .add_link('item', "/images/48184234")
                             .embed('item',
                                    Builder("/images/48184234")
                                    .set_property('width', 2048)
                                    .set_property('height', 1536)
                                    .add_link(
                                        'enclosure',
                                        "http://images.localhost/48184234.png",
                                        type="image/jpeg"))))
               .add_link('api:pictures',
                         "/products/254-rocket-skates/pictures/")
               .add_link('api:price', "/prices/537461",
                         title="$24.95 + $1.00 delivery")
               .embed('api:price',
                      Builder("/prices/537461")
                      .set_property("currency_code", "USD")
                      .set_property("currency_symbol", "$")
                      .set_property("advertised_price", "24.95")
                      .set_property("delivery_price", "1.00")
                      .set_property("total_price", "25.95"))
               ).as_object()

        doc = Document.from_object(obj, base_uri="http://localhost/")
        self.assertEqual(doc.properties['name'], "Rocket Skates")
        self.assertEqual(doc.links['/rels/reviews'].url(),
                         "http://localhost/products/254-rocket-skates/reviews")
        item = doc.embedded['/rels/pictures'].embedded['item'].embedded['item']
        self.assertEquals(item.properties['width'], 2048)
        self.assertEquals(item.links['enclosure'].type, "image/jpeg")


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.

import urlparse
from random import choice


class CurieCollection(dict):
    def __init__(self, expansions_cache={}):
        # The default value of expansions_cache is not a mistake. The
        # dictionary is deliberately shared because the mapping between
        # (template, rel) and expansion is valid everywhere.
        super(CurieCollection, self).__init__()
        self.expansions_cache = expansions_cache

    EXPANSIONS_CACHE_LEN_LIMIT = 128

    def expand(self, link):
        terms = link.split(':')

        if len(terms) < 2:
            return link

        key, value = terms

        if key not in self:
            return link

        template = self[key]
        memo_key = template.href, value

        if memo_key in self.expansions_cache:
            return self.expansions_cache[memo_key]

        # Prevent unbounded growth of the shared cache.
        if len(self.expansions_cache) >= self.EXPANSIONS_CACHE_LEN_LIMIT:
            del self.expansions_cache[choice(self.expansions_cache.keys())]

        result = self[key].url(rel=value)
        self.expansions_cache[memo_key] = result

        return result



#!/usr/bin/env python
# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.

from distutils.core import setup

VERSION = "0.5"

base_url = "http://github.com/wharris/dougrain/"

setup(
    name='dougrain',
    version=VERSION,
    description='HAL JSON parser and generator',
    author='Will Harris',
    author_email='will@greatlibrary.net',
    url=base_url,
    packages=['dougrain'],
    provides=['dougrain'],
    long_description=open("README.rst").read(),
    install_requires=['uritemplate >= 0.5.1'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Operating System :: POSIX',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



# Copyright (c) 2013 Will Harris
# See the file license.txt for copying permission.

import urlparse
import re
import uritemplate


def extract_variables(href):
    """Return a list of variable names used in a URI template."""

    patterns = [re.sub(r'\*|:\d+', '', pattern)
                for pattern in re.findall(r'{[\+#\./;\?&]?([^}]+)*}', href)]
    variables = []
    for pattern in patterns:
        for part in pattern.split(","):
            if not part in variables:
                variables.append(part)

    return variables


class Link(object):
    """Representation of a HAL link from a ``Document``.

    Constructors:

    - ``Link.from_object(o, base_uri=None)``:
        returns a new ``Link`` based on a JSON object.

    Public Instance Attributes:

    - ``href``: ``str`` containing the href of the link.
    - ``name``: ``str`` containing the name of the link. Absent if the link
                has no name.
    - ``title``: ``str`` containing the title of the link. Absent if the link
                 has no title.
    - ``type``: ``str`` containing the type of the link. Absent if the link
                does not specify a type.
    - ``profile``: ``str`` containing the profile URL reference of the link.
                   Absent if the link does not specify a profile.
    - ``hreflang``: ``str`` indicating the language of the target. Absent if
                    the link does not specify a language.
    - ``deprecation``: ``str`` indicating that the link is deprecated. The
                       value of the string should be a URL that provides
                       further infomation about the deprecation. Absent if the
                       link is not marked as deprecated.
    - ``variables``: ``list`` of names of template variables that may be
                     expanded for templated links. Empty if there are no
                     template variables.

    """
    def __init__(self, json_object, base_uri):
        self.o = json_object
        self.href = json_object['href']

        if 'name' in json_object:
            self.name = json_object['name']

        if 'title' in json_object:
            self.title = json_object['title']

        if 'type' in json_object:
            self.type = json_object['type']

        if 'profile' in json_object:
            self.profile = json_object['profile']

        if 'hreflang' in json_object:
            self.hreflang = json_object['hreflang']

        if 'deprecation' in json_object:
            self.deprecation = json_object['deprecation']

        self.is_templated = self.o.get('templated', False) is True

        if self.is_templated:
            self.variables = extract_variables(self.href)
        else:
            self.variables = []

        if base_uri is None:
            self.template = self.href
        else:
            self.template = urlparse.urljoin(base_uri, self.href)

    def url(self, **kwargs):
        """Returns a URL for the link with optional template expansion.

        If the link is marked as templated, the href will be expanded according
        to RFC6570, using template variables provided in the keyword arguments.
        If the href is a valid URI Template, but the link is not marked as
        templated, the href will not be expanded even if template variables are
        provided.

        """
        if self.is_templated:
            return uritemplate.expand(self.template, kwargs)
        else:
            return self.template

    def as_object(self):
        """Returns a dictionary representing the HAL JSON link."""
        return self.o

    def as_link(self):
        """Returns a ``Link`` to the same resource as this link.

        This method is trivial, but is provided for symmetry with ``Document``.

        """
        return self

    @classmethod
    def from_object(cls, o, base_uri):
        """Returns a new ``Link`` based on a JSON object or array.

        Arguments:

        - ``o``: a dictionary holding the deserializated JSON for the new
                 ``Link``, or a ``list`` of such documents.
        - ``base_uri``: optional URL used as the basis when expanding
                               relative URLs in the link.

        """
        if isinstance(o, list):
            if len(o) == 1:
                return cls.from_object(o[0], base_uri)

            return map(lambda x: cls.from_object(x, base_uri), o)

        return cls(o, base_uri)

    def __iter__(self):
        yield self

    def __repr__(self):
        if hasattr(self, 'name'):
            return "<Link %s=%r>" % (self.name, self.template)
        else:
            return "<Link %r>" % self.template

    def __eq__(self, other):
        return (isinstance(other, Link) and
                self.as_object() == other.as_object())
